"""Unit tests for reviewboard.licensing.views.LicensesView.

Version Added:
    8.0
"""

from __future__ import annotations

import json
from datetime import datetime, timezone as tz
from uuid import uuid4

import kgb
from django.urls import reverse
from django.utils import timezone
from djblets.features.testing import override_feature_check

from reviewboard.licensing.features import licensing_feature
from reviewboard.licensing.provider import BaseLicenseProvider
from reviewboard.licensing.registry import license_provider_registry
from reviewboard.licensing.tests.providers import (
    BasicTestsLicenseProvider,
    ExpiredTestsLicenseProvider,
)
from reviewboard.testing import TestCase


class LicenseViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for LicenseView.

    Version Added:
        8.0
    """

    fixtures = ['test_users']

    ######################
    # Instance variables #
    ######################

    #: A static timestamp for "now", to ease unit testing.
    now: datetime

    #: The list of license providers registered for tests.
    _license_providers: list[BaseLicenseProvider]

    @classmethod
    def setUpClass(cls) -> None:
        """Set up state for all unit tests in the class.

        This will store a static timestamp for "now" and a list of license
        providers to register for the tests.
        """
        super().setUpClass()

        license_providers = [
            BasicTestsLicenseProvider(),
            ExpiredTestsLicenseProvider(),
        ]

        cls._license_providers = license_providers
        cls.now = datetime(2025, 4, 21, 0, 0, 0, tzinfo=tz.utc)

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down state for the unit tests."""
        super().tearDownClass()

        cls._license_providers = []
        cls.now = None  # type: ignore

    def setUp(self) -> None:
        """Set up state for a unit test.

        This will register all the license providers and force the current
        timestamp for "now".
        """
        super().setUp()

        for license_provider in self._license_providers:
            license_provider_registry.register(license_provider)

        self.spy_on(timezone.now, op=kgb.SpyOpReturn(self.now))

    def tearDown(self) -> None:
        """Tear down state for the test.

        This will unregister all the license providers.
        """
        super().tearDown()

        for license_provider in self._license_providers:
            license_provider_registry.unregister(license_provider)

    @override_feature_check(licensing_feature, True)
    def test_get_as_anonymous(self) -> None:
        """Testing LicenseView.get as anonymous"""
        client = self.client
        response = client.get(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 302)

    @override_feature_check(licensing_feature, True)
    def test_get_as_non_admin(self) -> None:
        """Testing LicenseView.get as non-admin user"""
        client = self.client
        self.assertTrue(
            self.client.login(username='dopey', password='dopey'))

        response = client.get(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 302)

    @override_feature_check(licensing_feature, True)
    def test_get_as_admin(self) -> None:
        """Testing LicenseView.get as admin"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.get(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['license_entries'], [
            {
                'license_infos': [
                    {
                        'attrs': {
                            'actionTarget': 'basic-tests-provider:license1',
                            'actions': [
                                {
                                    'actionID': 'test',
                                    'label': 'Test',
                                    'url': 'https://example.com/license1/',
                                },
                            ],
                            'autoRenew': False,
                            'canUploadLicense': False,
                            'expiresDate': datetime(2025, 7, 30, 0, 0,
                                                    tzinfo=tz.utc),
                            'expiresSoon': False,
                            'gracePeriodDaysRemaining': 0,
                            'hardExpiresDate': datetime(2025, 7, 30, 0, 0,
                                                        tzinfo=tz.utc),
                            'isTrial': False,
                            'licenseID': 'license1',
                            'licensedTo': 'Test User',
                            'lineItems': [
                                {
                                    'content': 'Line 1',
                                    'icon': 'ink-i-info',
                                },
                                {
                                    'content': '<strong>Line 2</strong>',
                                    'contentIsHTML': True,
                                },
                            ],
                            'manageURL': 'https://example.com/license1/',
                            'noticeHTML': '',
                            'planID': None,
                            'planName': None,
                            'productName': 'Test Product',
                            'status': 'unlicensed',
                            'summary': 'Test Product is not licensed!',
                        },
                    },
                    {
                        'attrs': {
                            'actionTarget': 'basic-tests-provider:license2',
                            'actions': [
                                {
                                    'actionID': 'test',
                                    'label': 'Test',
                                    'url': 'https://example.com/license2/',
                                },
                            ],
                            'autoRenew': False,
                            'canUploadLicense': False,
                            'expiresDate': datetime(2025, 4, 26, 0, 0,
                                                    tzinfo=tz.utc),
                            'expiresSoon': True,
                            'gracePeriodDaysRemaining': 0,
                            'hardExpiresDate': datetime(2025, 4, 26, 0, 0,
                                                        tzinfo=tz.utc),
                            'isTrial': False,
                            'licenseID': 'license2',
                            'licensedTo': 'Test User',
                            'lineItems': [],
                            'manageURL': None,
                            'noticeHTML': '',
                            'planID': 'plan1',
                            'planName': 'Plan 1',
                            'productName': 'Test Product',
                            'status': 'licensed',
                            'summary': (
                                'License for Test Product Plan 1 '
                                'expires in 5 days'
                            ),
                        },
                    },
                ],
                'model': 'RB.License',
                'view': 'RB.LicenseView',
            },
            {
                'license_infos': [
                    {
                        'attrs': {
                            'actionTarget': 'expired-tests-provider:license1',
                            'actions': [],
                            'autoRenew': False,
                            'canUploadLicense': False,
                            'expiresDate': datetime(2025, 3, 2, 0, 0,
                                                    tzinfo=tz.utc),
                            'expiresSoon': False,
                            'gracePeriodDaysRemaining': 0,
                            'hardExpiresDate': datetime(2025, 3, 2, 0, 0,
                                                        tzinfo=tz.utc),
                            'isTrial': True,
                            'licenseID': 'license1',
                            'licensedTo': 'Test User',
                            'lineItems': [],
                            'manageURL': None,
                            'noticeHTML': '',
                            'planID': None,
                            'planName': None,
                            'productName': 'Test Product',
                            'status': 'hard-expired',
                            'summary': (
                                'Trial license for Test Product expired'
                            ),
                        },
                    },
                    {
                        'attrs': {
                            'actionTarget': 'expired-tests-provider:license2',
                            'actions': [],
                            'autoRenew': False,
                            'canUploadLicense': False,
                            'expiresDate': datetime(2025, 4, 19, 0, 0,
                                                    tzinfo=tz.utc),
                            'expiresSoon': False,
                            'gracePeriodDaysRemaining': 3,
                            'hardExpiresDate': datetime(2025, 4, 22, 0, 0,
                                                        tzinfo=tz.utc),
                            'isTrial': False,
                            'licenseID': 'license2',
                            'licensedTo': 'Test User',
                            'lineItems': [],
                            'manageURL': None,
                            'noticeHTML': (
                                'Your grace period is now active. Unless '
                                'renewed, Test Product will be disabled <time '
                                'class="timesince" dateTime="2025-04-22 '
                                '00:00:00+00:00"/>.'
                            ),
                            'planID': None,
                            'planName': None,
                            'productName': 'Test Product',
                            'status': 'expired-grace-period',
                            'summary': (
                                'License for Test Product expired and will '
                                'stop working in 3 days'
                            ),
                        },
                    },
                ],
                'model': 'RB.License',
                'view': 'RB.LicenseView',
            },
        ])

    @override_feature_check(licensing_feature, True)
    def test_post_as_anonymous(self) -> None:
        """Testing LicenseView.post as anonymous"""
        client = self.client
        response = client.post(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 302)

    @override_feature_check(licensing_feature, True)
    def test_post_as_non_admin(self) -> None:
        """Testing LicenseView.post as non-admin user"""
        client = self.client
        self.assertTrue(self.client.login(username='dopey', password='dopey'))
        response = client.post(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 302)

    @override_feature_check(licensing_feature, True)
    def test_post_with_no_action(self) -> None:
        """Testing LicenseView.post as admin with no action"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'))

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': 'Missing action data.',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_no_action_target(self) -> None:
        """Testing LicenseView.post as admin with no action_target"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'xxx',
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': 'Missing action target.',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_invalid_action_target(self) -> None:
        """Testing LicenseView.post as admin with invalid action_target"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'xxx',
            'action_target': 'xxx',
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': 'Invalid action target.',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_license_not_found(self) -> None:
        """Testing LicenseView.post as admin with license not found"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'xxx',
            'action_target': 'expired-tests-provider:xxx',
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': 'The license entry could not be found.',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_invalid_action(self) -> None:
        """Testing LicenseView.post as admin with invalid action"""
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'xxx',
            'action_target': 'expired-tests-provider:license1',
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': 'Unsupported license action "xxx".',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_invalid_action_data(
        self,
    ) -> None:
        """Testing LicenseView.post as admin with invalid action_data"""
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
            'action_data': (
                '{'
                '"check_request_data": "abc123",'
                '"check_response_data": xxx'
                '}'
            ),
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': (
                'Invalid action data: Expecting value: line 1 column 56 '
                '(char 55).'
            ),
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_license_update_check(self) -> None:
        """Testing LicenseView.post as admin with action="license-update-check"
        """
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'license-update-check',
            'action_target': 'basic-tests-provider:license1',
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {
            'canCheck': True,
            'checkStatusURL': 'https://example.com/license1/check/',
            'credentials': None,
            'data': {
                'license_id': 'license1',
                'something': 'special',
                'version': '1.0',
            },
            'headers': None,
            'sessionToken': 'abc123',
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_license_update_check_not_supported(self) -> None:
        """Testing LicenseView.post as admin with action="license-update-check"
        not supported
        """
        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'license-update-check',
            'action_target': 'expired-tests-provider:license1',
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {
            'canCheck': False,
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_process_license_update_and_applied(self) -> None:
        """Testing LicenseView.post as admin with
        action="process-license-update" and new license data applied
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
            'action_data': json.dumps({
                'check_request_data': {
                    'license_id': 'license1',
                    'something': 'special',
                    'version': '1.0',
                },
                'check_response_data': {
                    'updated': True,
                    'version': '1.0',
                },
            }),
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {
            'status': 'applied',
            'license_infos': {
                'license1': {
                    'actionTarget': 'basic-tests-provider:license2',
                    'actions': [
                        {
                            'actionID': 'test',
                            'label': 'Test',
                            'url': 'https://example.com/license2/',
                        },
                    ],
                    'autoRenew': False,
                    'canUploadLicense': False,
                    'expiresDate': '2026-04-21T00:00:00Z',
                    'expiresSoon': False,
                    'gracePeriodDaysRemaining': 0,
                    'hardExpiresDate': '2026-04-21T00:00:00Z',
                    'isTrial': False,
                    'licenseID': 'license2',
                    'licensedTo': 'Test User',
                    'lineItems': [],
                    'manageURL': None,
                    'noticeHTML': '',
                    'planID': 'smpbpe1',
                    'planName': 'Super Mega Power Bundle Pro Enterprise',
                    'productName': 'Test Product',
                    'status': 'licensed',
                    'summary': (
                        'License for Test Product Super Mega Power Bundle '
                        'Pro Enterprise is active'
                    ),
                },
            },
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_process_license_update_and_has_latest(self) -> None:
        """Testing LicenseView.post as admin with
        action="process-license-update" and already has latest data
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
            'action_data': json.dumps({
                'check_request_data': {
                    'license_id': 'license1',
                    'something': 'special',
                    'version': '1.0',
                },
                'check_response_data': {
                    'latest': True,
                    'version': '1.0',
                },
            }),
        })

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {
            'status': 'has-latest',
            'license_infos': {
                'license1': {
                    'actionTarget': 'basic-tests-provider:license2',
                    'actions': [
                        {
                            'actionID': 'test',
                            'label': 'Test',
                            'url': 'https://example.com/license2/',
                        },
                    ],
                    'autoRenew': False,
                    'canUploadLicense': False,
                    'expiresDate': '2025-04-26T00:00:00Z',
                    'expiresSoon': True,
                    'gracePeriodDaysRemaining': 0,
                    'hardExpiresDate': '2025-04-26T00:00:00Z',
                    'isTrial': False,
                    'licenseID': 'license2',
                    'licensedTo': 'Test User',
                    'lineItems': [],
                    'manageURL': None,
                    'noticeHTML': '',
                    'planID': 'plan1',
                    'planName': 'Plan 1',
                    'productName': 'Test Product',
                    'status': 'licensed',
                    'summary': (
                        'License for Test Product Plan 1 expires in '
                        '5 days'
                    ),
                },
            },
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_process_license_update_and_error(self) -> None:
        """Testing LicenseView.post as admin with
        action="process-license-update" and error
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
            'action_data': json.dumps({
                'check_request_data': {
                    'license_id': 'license1',
                    'something': 'special',
                    'version': '1.0',
                },
                'check_response_data': {
                    'latest': True,
                    'version': '2.0',
                },
            }),
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': (
                'Error processing license update: Invalid version from '
                'the licensing server! Oh no!'
            ),
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_process_license_update_no_check_data(self) -> None:
        """Testing LicenseView.post as admin with
        action="process-license-update" and missing check_request_data
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': (
                f'Missing check_request_data value for license check. This '
                f'may be an internal error or an issue with the licensing '
                f'server. Check the Review Board server logs for more '
                f'information (error ID {trace_id}).'
            ),
        })

    @override_feature_check(licensing_feature, True)
    def test_post_with_process_license_update_no_response_data(self) -> None:
        """Testing LicenseView.post as admin with
        action="process-license-update" and missing check_response_data
        """
        trace_id = '00000000-0000-0000-0000-000000000001'
        self.spy_on(uuid4, op=kgb.SpyOpReturn(trace_id))

        client = self.client
        self.assertTrue(client.login(username='admin', password='admin'))

        response = client.post(reverse('admin-licenses'), {
            'action': 'process-license-update',
            'action_target': 'basic-tests-provider:license2',
            'action_data': json.dumps({
                'check_request_data': 'abc123',
            }),
        })

        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {
            'error': (
                f'Missing check_response_data value for license check. This '
                f'may be an internal error or an issue with the licensing '
                f'server. Check the Review Board server logs for more '
                f'information (error ID {trace_id}).'
            ),
        })



"""Forms for uploading diffs."""

from functools import partial

from dateutil.parser import isoparse
from django import forms
from django.core.exceptions import ValidationError
from django.utils.encoding import force_str
from django.utils.translation import gettext, gettext_lazy as _

from reviewboard.diffviewer.commit_utils import (deserialize_validation_info,
                                                 get_file_exists_in_history)
from reviewboard.diffviewer.differ import DiffCompatVersion
from reviewboard.diffviewer.diffutils import check_diff_size
from reviewboard.diffviewer.filediff_creator import create_filediffs
from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.diffviewer.validators import (COMMIT_ID_LENGTH,
                                               validate_commit_id)


class BaseCommitValidationForm(forms.Form):
    """A form mixin for handling validation metadata for commits."""

    validation_info = forms.CharField(
        label=_('Validation metadata'),
        help_text=_('Validation metadata generated by the diff commit '
                    'validation resource.'),
        widget=forms.HiddenInput,
        required=False)

    def clean_validation_info(self):
        """Clean the validation_info field.

        This method ensures that if the field is supplied that it parses as
        base64-encoded JSON.

        Returns:
            dict:
            The parsed validation information.

        Raises:
            django.core.exceptions.ValidationError:
                The value could not be parsed.
        """
        validation_info = self.cleaned_data.get('validation_info', '').strip()

        if not validation_info:
            return {}

        try:
            return deserialize_validation_info(validation_info)
        except (TypeError, ValueError) as e:
            raise ValidationError(
                gettext(
                    'Could not parse validation info "%(validation_info)s": '
                    '%(exc)s'
                ) % {
                    'exc': e,
                    'validation_info': validation_info,
                })


class UploadCommitForm(BaseCommitValidationForm):
    """The form for uploading a diff and creating a DiffCommit."""

    diff = forms.FileField(
        label=_('Diff'),
        help_text=_('The new diff to upload.'))

    parent_diff = forms.FileField(
        label=_('Parent diff'),
        help_text=_('An optional diff that the main diff is based on. '
                    'This is usually used for distributed revision control '
                    'systems (Git, Mercurial, etc.).'),
        required=False)

    commit_id = forms.CharField(
        label=_('Commit ID'),
        help_text=_('The ID of this commit.'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id])

    parent_id = forms.CharField(
        label=_('Parent commit ID'),
        help_text=_('The ID of the parent commit.'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id])

    commit_message = forms.CharField(
        label=_('Description'),
        help_text=_('The commit message.'))

    author_name = forms.CharField(
        label=_('Author name'),
        help_text=_('The name of the author of this commit.'),
        max_length=DiffCommit.NAME_MAX_LENGTH)

    author_email = forms.CharField(
        label=_('Author e-mail address'),
        help_text=_('The e-mail address of the author of this commit.'),
        max_length=DiffCommit.EMAIL_MAX_LENGTH,
        widget=forms.EmailInput)

    author_date = forms.CharField(
        label=_('Author date'),
        help_text=_('The date and time this commit was authored.'))

    committer_name = forms.CharField(
        label=_('Committer name'),
        help_text=_('The name of the committer of this commit.'),
        max_length=DiffCommit.NAME_MAX_LENGTH,
        required=True)

    committer_email = forms.CharField(
        label=_('Committer e-mail address'),
        help_text=_('The e-mail address of the committer of this commit.'),
        max_length=DiffCommit.EMAIL_MAX_LENGTH,
        widget=forms.EmailInput,
        required=True)

    committer_date = forms.CharField(
        label=_('Committer date'),
        help_text=_('The date and time this commit was committed.'),
        required=True)

    def __init__(self, diffset, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            diffset (reviewboard.diffviewer.models.diffset.DiffSet):
                The DiffSet to attach the created DiffCommit to.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super(UploadCommitForm, self).__init__(*args, **kwargs)

        if not diffset.repository.scmtool_class.commits_have_committer:
            del self.fields['committer_date']
            del self.fields['committer_email']
            del self.fields['committer_name']

        self.diffset = diffset
        self.request = request

    def create(self):
        """Create the DiffCommit.

        Returns:
            reviewboard.diffviewer.models.diffcommit.DiffCommit:
            The created DiffCommit.
        """
        assert self.is_valid()

        return DiffCommit.objects.create_from_upload(
            request=self.request,
            validation_info=self.cleaned_data['validation_info'],
            diffset=self.diffset,
            repository=self.diffset.repository,
            base_commit_id=self.diffset.base_commit_id,
            diff_file=self.cleaned_data['diff'],
            parent_diff_file=self.cleaned_data.get('parent_diff'),
            commit_message=self.cleaned_data['commit_message'],
            commit_id=self.cleaned_data['commit_id'],
            parent_id=self.cleaned_data['parent_id'],
            author_name=self.cleaned_data['author_name'],
            author_email=self.cleaned_data['author_email'],
            author_date=self.cleaned_data['author_date'],
            committer_name=self.cleaned_data.get('committer_name'),
            committer_email=self.cleaned_data.get('committer_email'),
            committer_date=self.cleaned_data.get('committer_date'))

    def clean(self):
        """Clean the form.

        Returns:
            dict:
            The cleaned form data.

        Raises:
            django.core.exceptions.ValidationError:
                The form data was not valid.
        """
        super(UploadCommitForm, self).clean()

        if self.diffset.history_id is not None:
            # A diffset will have a history attached if and only if it has been
            # published, in which case we cannot attach further commits to it.
            raise ValidationError(gettext(
                'Cannot upload commits to a published diff.'))

        if (self.diffset.commit_count and
            'validation_info' not in self.cleaned_data and
            'validation_info' not in self.errors):
            # If validation_info is present in `errors`, it will not be in
            # self.cleaned_data. We do not want to report it missing if it
            # failed validation for another reason.
            self._errors['validation_info'] = self.error_class([
                self.fields['validation_info'].error_messages['required'],
            ])

        return self.cleaned_data

    def clean_author_date(self):
        """Parse the date and time in the author_date field.

        Returns:
            datetime.datetime:
            The parsed date and time.
        """
        try:
            return isoparse(self.cleaned_data['author_date'])
        except ValueError:
            raise ValidationError(gettext(
                'This date must be in ISO 8601 format.'))

    def clean_committer_date(self):
        """Parse the date and time in the committer_date field.

        Returns:
            datetime.datetime:
            The parsed date and time.
        """
        try:
            return isoparse(self.cleaned_data['committer_date'])
        except ValueError:
            raise ValidationError(gettext(
                'This date must be in ISO 8601 format.'))


class UploadDiffForm(forms.Form):
    """The form for uploading a diff and creating a DiffSet."""

    path = forms.FileField(
        label=_('Diff'),
        help_text=_('The new diff to upload.'))

    parent_diff_path = forms.FileField(
        label=_('Parent Diff'),
        help_text=_('An optional diff that the main diff is based on. '
                    'This is usually used for distributed revision control '
                    'systems (Git, Mercurial, etc.).'),
        required=False)

    basedir = forms.CharField(
        label=_('Base Directory'),
        help_text=_('The absolute path in the repository the diff was '
                    'generated in.'))

    base_commit_id = forms.CharField(
        label=_('Base Commit ID'),
        help_text=_('The ID/revision this change is built upon.'),
        required=False)

    def __init__(self, repository, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository the diff will be uploaded against.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwrgs (dict):
                Additional keyword arguments.
        """
        super(UploadDiffForm, self).__init__(*args, **kwargs)

        self.repository = repository
        self.request = request

        if repository.diffs_use_absolute_paths:
            # This SCMTool uses absolute paths, so there's no need to ask
            # the user for the base directory.
            del(self.fields['basedir'])

    def clean_base_commit_id(self):
        """Clean the ``base_commit_id`` field.

        Returns:
            unicode:
            The ``base_commit_id`` field stripped of leading and trailing
            whitespace, or ``None`` if that value would be empty.
        """
        return self.cleaned_data['base_commit_id'].strip() or None

    def clean_basedir(self):
        """Clean the ``basedir`` field.

        Returns:
            unicode:
            The basedir field as a unicode string with leading and trailing
            whitespace removed.
        """
        if self.repository.diffs_use_absolute_paths:
            return ''

        return force_str(self.cleaned_data['basedir'].strip())

    def create(self, diffset_history=None):
        """Create the DiffSet.

        Args:
            diffset_history (reviewboard.diffviewer.models.diffset_history.
                             DiffSetHistory):
                The DiffSet history to attach the created DiffSet to.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The created DiffSet.
        """
        assert self.is_valid()

        return DiffSet.objects.create_from_upload(
            repository=self.repository,
            diffset_history=diffset_history,
            diff_file=self.cleaned_data['path'],
            parent_diff_file=self.cleaned_data.get('parent_diff_path'),
            basedir=self.cleaned_data.get('basedir', ''),
            base_commit_id=self.cleaned_data['base_commit_id'],
            request=self.request)


class ValidateCommitForm(BaseCommitValidationForm):
    """A form for validating of DiffCommits."""

    diff = forms.FileField(
        label=_('Diff'),
        help_text=_('The new diff to upload.'))

    parent_diff = forms.FileField(
        label=_('Parent diff'),
        help_text=_('An optional diff that the main diff is based on. '
                    'This is usually used for distributed revision control '
                    'systems (Git, Mercurial, etc.).'),
        required=False)

    commit_id = forms.CharField(
        label=_('Commit ID'),
        help_text=_('The ID of this commit.'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id])

    parent_id = forms.CharField(
        label=_('Parent commit ID'),
        help_text=_('The ID of the parent commit.'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id])

    base_commit_id = forms.CharField(
        label=_('Base commit ID'),
        help_text=_('The base commit ID that the commits are based off of.'),
        required=False)

    def __init__(self, repository, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository against which the diff is being validated.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Additional positional arguments to pass to the base
                class initializer.

            **kwargs (dict):
                Additional keyword arguments to pass to the base class
                initializer.
        """
        super(ValidateCommitForm, self).__init__(*args, **kwargs)

        self.repository = repository
        self.request = request

    def clean(self):
        """Clean the form.

        Returns:
            dict:
            The cleaned form data.

        Raises:
            django.core.exceptions.ValidationError:
                The form data was not valid.
        """
        super(ValidateCommitForm, self).clean()

        validation_info = self.cleaned_data.get('validation_info')

        if validation_info:
            errors = []

            parent_id = self.cleaned_data.get('parent_id')
            commit_id = self.cleaned_data.get('commit_id')

            if commit_id and commit_id in validation_info:
                errors.append(gettext('This commit was already validated.'))
            elif parent_id and parent_id not in validation_info:
                errors.append(gettext('The parent commit was not validated.'))

            if errors:
                self._errors['validation_info'] = self.error_class(errors)
                self.cleaned_data.pop('validation_info')

        return self.cleaned_data

    def validate_diff(self):
        """Validate the DiffCommit.

        This will attempt to parse the given diff (and optionally parent
        diff) into :py:class:`FileDiffs
        <reviewboard.diffviewer.models.filediff.FileDiff>`. This will not
        result in anything being committed to the database.

        Returns:
            tuple:
            A 2-tuple containing the following:

            * A list of the created FileDiffs.
            * A list of the parent FileDiffs, or ``None``.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                The diff could not be parsed.

            reviewboard.diffviewer.errors.DiffTooBigError:
                The diff was too big.

            reviewboard.diffviewer.errors.EmptyDiffError:
                The diff did not contain any changes.

            reviewboard.scmtools.errors.FileNotFoundError:
                A file was not found in the repository.

            reviewboard.scmtools.errors.SCMError:
                An error occurred within the SCMTool.
        """
        assert self.is_valid()

        diff_file = self.cleaned_data['diff']
        parent_diff_file = self.cleaned_data.get('parent_diff')
        validation_info = self.cleaned_data.get('validation_info')

        check_diff_size(diff_file, parent_diff_file)

        if parent_diff_file:
            parent_diff_file_contents = parent_diff_file.read()
        else:
            parent_diff_file_contents = None

        base_commit_id = self.cleaned_data['base_commit_id']

        diffset = DiffSet(name='diff',
                          revision=0,
                          basedir='',
                          repository=self.repository,
                          diffcompat=DiffCompatVersion.DEFAULT,
                          base_commit_id=base_commit_id)

        get_file_exists = partial(get_file_exists_in_history,
                                  validation_info or {},
                                  self.repository,
                                  self.cleaned_data['parent_id'])

        return create_filediffs(
            diff_file_contents=diff_file.read(),
            parent_diff_file_contents=parent_diff_file_contents,
            repository=self.repository,
            basedir='',
            base_commit_id=base_commit_id,
            get_file_exists=get_file_exists,
            diffset=diffset,
            request=self.request,
            diffcommit=None,
            validate_only=True)



"""Unit tests for reviewboard.themes.ui.registry.UIThemeRegistry.

Version Added:
    7.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from reviewboard.testing import TestCase
from reviewboard.themes.ui.default import DarkUITheme, SystemUITheme
from reviewboard.themes.ui.registry import UIThemeRegistry

if TYPE_CHECKING:
    from typelets.json import JSONDict


class UIThemeRegistryTests(TestCase):
    """Unit tests for UIThemeRegistry.

    Version Added:
        7.0
    """

    ######################
    # Instance variables #
    ######################

    #: The registry used in tests.
    registry: UIThemeRegistry

    def setUp(self) -> None:
        super().setUp()

        self.registry = UIThemeRegistry()

    def test_get_theme(self) -> None:
        """Testing UIThemeRegistry.get_theme"""
        self.assertIsInstance(self.registry.get_theme('dark'), DarkUITheme)

    def test_get_theme_with_default(self) -> None:
        """Testing UIThemeRegistry.get_theme with ID 'default'"""
        self.assertIsInstance(self.registry.get_theme('default'),
                              SystemUITheme)

    def test_get_default_theme_id(self) -> None:
        """Testing UIThemeRegistry.get_default_theme_id"""
        self.assertEqual(self.registry.get_default_theme_id(), 'system')

    def test_get_default_theme_id_with_siteconfig(self) -> None:
        """Testing UIThemeRegistry.get_default_theme_id with custom default
        in siteconfig
        """
        siteconfig_settings: JSONDict = {
            'default_ui_theme': 'system',
        }

        with self.siteconfig_settings(siteconfig_settings):
            self.assertEqual(self.registry.get_default_theme_id(),
                             'system')

    def test_get_default_theme_id_with_siteconfig_not_found(self) -> None:
        """Testing UIThemeRegistry.get_default_theme_id with custom default
        in siteconfig not found in registry
        """
        siteconfig_settings: JSONDict = {
            'default_ui_theme': 'xxx',
        }

        with self.siteconfig_settings(siteconfig_settings):
            self.assertEqual(self.registry.get_default_theme_id(), 'system')



"""Diffviewer features."""

from django.utils.translation import gettext_lazy as _
from djblets.features import Feature, FeatureLevel


class DVCSFeature(Feature):
    """A feature for DVCS support.

    With this enabled, the DVCS API is enabled and review requests can be
    created with multiple commits.
    """

    feature_id = 'diffviewer.dvcs'
    name = _('DVCS Support')
    level = FeatureLevel.STABLE
    summary = _('Support for reviewing multiple commits.')


dvcs_feature = DVCSFeature()



"""Unit tests for reviewboard.admin.checks."""

from django.conf import settings
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin import checks
from reviewboard.testing.testcase import TestCase


class ChecksTests(TestCase):
    """Unit tests for reviewboard.admin.checks."""

    def setUp(self):
        super(ChecksTests, self).setUp()

        self.old_media_root = settings.MEDIA_ROOT

    def tearDown(self):
        super(ChecksTests, self).tearDown()

        # Make sure we don't break further tests by resetting this fully.
        checks.reset_check_cache()

        # If test_manual_updates_bad_upload failed in the middle, it could
        # neglect to fix the MEDIA_ROOT, which will break a bunch of future
        # tests. Make sure it's always what we expect.
        settings.MEDIA_ROOT = self.old_media_root
        siteconfig = SiteConfiguration.objects.get_current()
        siteconfig.set('site_media_root', self.old_media_root)
        siteconfig.save()

    def test_check_updates_required_with_no_updates(self):
        """Testing check_updates_required with valid configuration"""
        # NOTE: This is assuming the install is fine. It should be given
        #       that we set things like the uploaded path correctly to
        #       a known good directory before starting unit tests.
        updates_required = checks.check_updates_required()

        self.assertEqual(len(updates_required), 0)

    def test_check_updates_required_with_bad_upload(self):
        """Testing check_updates_required with a bad upload directory"""
        settings.MEDIA_ROOT = "/"
        checks.reset_check_cache()

        with self.siteconfig_settings({'site_media_root': '/'},
                                      reload_settings=False):
            updates_required = checks.check_updates_required()
            self.assertEqual(len(updates_required), 2)

            url, data = updates_required[0]
            self.assertEqual(url,
                             'admin/manual-updates/media-upload-dir.html')

            response = self.client.get('/dashboard/')
            self.assertEqual(response.status_code, 200)
            self.assertTemplateUsed(response,
                                    'admin/manual_updates_required.html')

        settings.MEDIA_ROOT = self.old_media_root

        # Make sure that the site works again once the media root is fixed.
        with self.siteconfig_settings({'site_media_root': self.old_media_root},
                                      reload_settings=False):
            response = self.client.get('/dashboard/')
            self.assertTemplateNotUsed(response,
                                       'admin/manual_updates_required.html')



"""Unit tests for reviewboard.actions.renderers.MenuActionGroupRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.renderers import MenuActionGroupRenderer
from reviewboard.actions.tests.base import TestActionsRegistry, TestMenuAction
from reviewboard.deprecation import RemovedInReviewBoard10_0Warning
from reviewboard.testing import TestCase


class MenuActionGroupRendererTests(TestCase):
    """Unit tests for MenuActionGroupRenderer.

    Version Added:
        8.0
    """

    def test_render(self) -> None:
        """Testing MenuActionGroupRenderer.render"""
        action = TestMenuAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="menuitem">
             <span id="action-review-request-menu-action"
                   role="presentation">
              <a aria-label="Test Menu"
                 href="#"
                 role="presentation">
               Test Menu
               <span class="ink-i-dropdown"/>
              </a>
              <div hidden style="display: none;"></div>
             </span>
            </li>
            """)

    def test_render_with_icon_only(self) -> None:
        """Testing MenuActionGroupRenderer.render with icon only"""
        class MyTestMenuAction(TestMenuAction):
            icon_class = 'my-icon'
            label = ''
            verbose_label = 'My menu'

        action = MyTestMenuAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action -is-icon"
                role="menuitem">
             <span id="action-review-request-menu-action"
                   role="presentation">
              <a aria-label="My menu"
                 href="#"
                 role="presentation">
               <span class="my-icon"></span>
              </a>
              <div hidden style="display: none;"></div>
             </span>
            </li>
            """)

    def test_render_with_action_template(self) -> None:
        """Testing MenuActionGroupRenderer.render with action.template_name
        set
        """
        class MyAction(TestMenuAction):
            template_name = 'actions/button_action.html'

        # We already test for the deprecation warning message in the action
        # tests. This just suppresses warning output for the test run.
        with self.assertWarns(RemovedInReviewBoard10_0Warning):
            action = MyAction()

        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action" role="presentation">
             <button aria-label="Test Menu"
                     class="ink-c-button"
                     id="action-review-request-menu-action"
                     type="button">
              <label class="ink-c-button__label">
               Test Menu
              </label>
             </button>
            </li>
            """)

    def test_render_js(self) -> None:
        """Testing MenuActionGroupRenderer.render_js"""
        action = TestMenuAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = MenuActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        js = renderer.render_js(request=request,
                                context=context)

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.MenuActionView({
                "attachmentPointID": "review-request",
                el:  $('#action-review-request-menu-action'),
                model: page.getAction("menu-action"),
            }));
            """)



"""Unit tests for reviewboard.admin.actions.BaseAdminSidebarManageItemAction.

Version Added:
    8.0
"""

from __future__ import annotations

from django.contrib.auth.models import User
from django.urls import NoReverseMatch

from reviewboard.admin.actions import BaseAdminSidebarManageItemAction
from reviewboard.testing.testcase import TestCase


class BaseAdminSidebarManageItemActionTests(TestCase):
    """Unit tests for BaseAdminSidebarManageItemAction.

    Version Added:
        8.0
    """

    def test_get_add_item_url_with_result(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_add_item_url with
        add_item_url_name
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'
            add_item_url_name = 'admin:auth_user_add'

        action = MyManageItemAction()

        self.assertEqual(action.get_add_item_url(),
                         '/admin/db/auth/user/add/')

    def test_get_add_item_url_with_unset(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_add_item_url with
        add_item_url_name not set
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'

        action = MyManageItemAction()

        self.assertIsNone(action.get_add_item_url())

    def test_get_add_item_url_with_error(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_add_item_url with
        exception
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'
            add_item_url_name = 'xxx-bad-url'

        action = MyManageItemAction()

        message = (
            "Reverse for 'xxx-bad-url' not found. 'xxx-bad-url' is not "
            "a valid view function or pattern name."
        )

        with self.assertRaisesMessage(NoReverseMatch, message):
            action.get_add_item_url()

    def test_get_item_count_with_result(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_item_count with
        item_queryset
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'
            item_queryset = User.objects.all()

        self.create_user(username='user1')
        self.create_user(username='user2')

        action = MyManageItemAction()

        self.assertEqual(action.get_item_count(), 2)

    def test_get_item_count_with_result_0(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_item_count with
        item_queryset and no items
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'
            item_queryset = User.objects.all()

        action = MyManageItemAction()

        self.assertEqual(action.get_item_count(), 0)

    def test_get_item_count_with_unset(self) -> None:
        """Testing BaseAdminSidebarManageItemAction.get_item_count with
        item_queryset unset
        """
        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'

        self.create_user(username='user1')

        action = MyManageItemAction()

        self.assertIsNone(action.get_item_count())



from kgb import SpyAgency

from reviewboard.hostingsvcs.utils.paginator import (APIPaginator,
                                                     InvalidPageError,
                                                     ProxyPaginator)
from reviewboard.testing import TestCase


class DummyAPIPaginator(APIPaginator):
    start_query_param = 'start'
    per_page_query_param = 'per-page'

    def fetch_url(self, url):
        return {
            'data': [1, 2, 3],
            'headers': {},
        }


class DummyMultiPageAPIPaginator(APIPaginator):
    def fetch_url(self, url):
        if url == 'http://example.com/':
            return {
                'data': ['a', 'b', 'c'],
                'per_page': 3,
                'total_count': 8,
                'next_url': 'http://example.com/?page=2',
            }
        elif url == 'http://example.com/?page=2':
            return {
                'data': ['d', 'e', 'f'],
                'per_page': 3,
                'total_count': 8,
                'prev_url': 'http://example.com/?page=1',
                'next_url': 'http://example.com/?page=3',
            }
        elif url == 'http://example.com/?page=3':
            return {
                'data': ['g', 'h'],
                'per_page': 3,
                'total_count': 8,
                'prev_url': 'http://example.com/?page=2',
            }
        else:
            raise AssertionError('Unexpected URL %s' % url)


class BasePaginatorTests(SpyAgency, TestCase):
    """Unit tests for BasePaginator."""

    def test_iter_pages(self):
        """Testing BasePaginator.iter_pages"""
        paginator = DummyMultiPageAPIPaginator(client=None,
                                               url='http://example.com/')
        pages = list(paginator.iter_pages())

        self.assertEqual(len(pages), 3)
        self.assertEqual(pages[0], ['a', 'b', 'c'])
        self.assertEqual(pages[1], ['d', 'e', 'f'])
        self.assertEqual(pages[2], ['g', 'h'])

    def test_iter_pages_with_max_pages(self):
        """Testing BasePaginator.iter_pages with max_pages"""
        self.spy_on(DummyMultiPageAPIPaginator.fetch_url)

        paginator = DummyMultiPageAPIPaginator(client=None,
                                               url='http://example.com/')

        pages = list(paginator.iter_pages(max_pages=2))

        self.assertEqual(len(pages), 2)
        self.assertEqual(pages[0], ['a', 'b', 'c'])
        self.assertEqual(pages[1], ['d', 'e', 'f'])
        self.assertEqual(len(DummyMultiPageAPIPaginator.fetch_url.calls), 2)

    def test_iter_items(self):
        """Testing BasePaginator.iter_items"""
        paginator = DummyMultiPageAPIPaginator(client=None,
                                               url='http://example.com/')

        self.assertEqual(list(paginator.iter_items()),
                         ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'])

    def test_iter_items_with_max_pages(self):
        """Testing BasePaginator.iter_items with max_pages"""
        self.spy_on(DummyMultiPageAPIPaginator.fetch_url)

        paginator = DummyMultiPageAPIPaginator(client=None,
                                               url='http://example.com/')

        self.assertEqual(list(paginator.iter_items(max_pages=2)),
                         ['a', 'b', 'c', 'd', 'e', 'f'])
        self.assertEqual(len(DummyMultiPageAPIPaginator.fetch_url.calls), 2)

    def test_iter(self):
        """Testing BasePaginator.__iter__"""
        paginator = DummyMultiPageAPIPaginator(client=None,
                                               url='http://example.com/')
        pages = list(iter(paginator))

        self.assertEqual(len(pages), 3)
        self.assertEqual(pages[0], ['a', 'b', 'c'])
        self.assertEqual(pages[1], ['d', 'e', 'f'])
        self.assertEqual(pages[2], ['g', 'h'])


class APIPaginatorTests(TestCase):
    """Tests for APIPaginator."""

    def test_construct_initial_load(self):
        """Testing APIPaginator construction performs initial load"""
        paginator = DummyAPIPaginator(
            client=None,
            url='http://example.com',
            start=10)
        self.assertEqual(paginator.page_data, [1, 2, 3])

    def test_construct_with_start(self):
        """Testing APIPaginator construction with start=<value>"""
        paginator = DummyAPIPaginator(
            client=None,
            url='http://example.com/api/list/?foo=1',
            start=10)
        query_params = paginator.request_kwargs['query']

        self.assertEqual(query_params['start'], 10)

    def test_construct_with_per_page(self):
        """Testing APIPaginator construction with per_page=<value>"""
        paginator = DummyAPIPaginator(
            client=None,
            url='http://example.com/api/list/?foo=1',
            per_page=10)
        query_params = paginator.request_kwargs['query']

        self.assertEqual(query_params['per-page'], 10)

    def test_extract_page_info(self):
        """Testing APIPaginator page information extraction"""
        class PageInfoAPIPaginator(APIPaginator):
            def fetch_url(self, url):
                return {
                    'data': ['a', 'b', 'c'],
                    'headers': {
                        'Foo': 'Bar',
                    },
                    'per_page': 10,
                    'total_count': 100,
                    'prev_url': 'http://example.com/?page=1',
                    'next_url': 'http://example.com/?page=3',
                }

        paginator = PageInfoAPIPaginator(client=None,
                                         url='http://example.com/')

        self.assertEqual(paginator.page_data, ['a', 'b', 'c'])
        self.assertEqual(paginator.page_headers['Foo'], 'Bar')
        self.assertEqual(paginator.per_page, 10)
        self.assertEqual(paginator.total_count, 100)
        self.assertEqual(paginator.prev_url, 'http://example.com/?page=1')
        self.assertEqual(paginator.next_url, 'http://example.com/?page=3')

    def test_prev(self):
        """Testing APIPaginator.prev"""
        prev_url = 'http://example.com/?page=1'

        paginator = DummyAPIPaginator(client=None,
                                      url='http://example.com')
        paginator.prev_url = prev_url

        self.assertTrue(paginator.has_prev)
        self.assertFalse(paginator.has_next)

        data = paginator.prev()

        self.assertEqual(data, [1, 2, 3])
        self.assertEqual(paginator.url, prev_url)

    def test_prev_without_prev_page(self):
        """Testing APIPaginator.prev without a previous page"""
        paginator = DummyAPIPaginator(client=None,
                                      url='http://example.com')
        url = paginator.url

        self.assertFalse(paginator.has_prev)
        self.assertRaises(InvalidPageError, paginator.prev)
        self.assertEqual(paginator.url, url)

    def test_next(self):
        """Testing APIPaginator.next"""
        next_url = 'http://example.com/?page=3'

        paginator = DummyAPIPaginator(client=None,
                                      url='http://example.com')
        paginator.next_url = next_url

        self.assertFalse(paginator.has_prev)
        self.assertTrue(paginator.has_next)

        data = paginator.next()

        self.assertEqual(data, [1, 2, 3])
        self.assertEqual(paginator.url, next_url)

    def test_next_without_next_page(self):
        """Testing APIPaginator.next without a next page"""
        paginator = DummyAPIPaginator(client=None,
                                      url='http://example.com')
        url = paginator.url

        self.assertFalse(paginator.has_next)
        self.assertRaises(InvalidPageError, paginator.next)
        self.assertEqual(paginator.url, url)


class ProxyPaginatorTests(TestCase):
    """Tests for ProxyPaginator."""

    def setUp(self):
        self.paginator = DummyAPIPaginator(client=None,
                                           url='http://example.com')
        self.proxy = ProxyPaginator(self.paginator)

    def test_has_prev(self):
        """Testing ProxyPaginator.has_prev"""
        self.assertFalse(self.proxy.has_prev)

        self.paginator.prev_url = 'http://example.com/?start=1'
        self.assertTrue(self.proxy.has_prev)

    def test_has_next(self):
        """Testing ProxyPaginator.has_next"""
        self.assertFalse(self.proxy.has_next)

        self.paginator.next_url = 'http://example.com/?start=2'
        self.assertTrue(self.proxy.has_next)

    def test_per_page(self):
        """Testing ProxyPaginator.per_page"""
        self.paginator.per_page = 10
        self.assertEqual(self.proxy.per_page, 10)

    def test_total_count(self):
        """Testing ProxyPaginator.total_count"""
        self.paginator.total_count = 100
        self.assertEqual(self.proxy.total_count, 100)

    def test_prev(self):
        """Testing ProxyPaginator.prev"""
        prev_url = 'http://example.com/?page=1'

        self.paginator.prev_url = prev_url

        self.assertTrue(self.proxy.has_prev)
        self.assertFalse(self.proxy.has_next)

        data = self.proxy.prev()

        self.assertEqual(data, [1, 2, 3])
        self.assertEqual(self.paginator.url, prev_url)

    def test_next(self):
        """Testing ProxyPaginator.next"""
        next_url = 'http://example.com/?page=3'

        self.paginator.next_url = next_url

        self.assertFalse(self.proxy.has_prev)
        self.assertTrue(self.proxy.has_next)

        data = self.proxy.next()

        self.assertEqual(data, [1, 2, 3])
        self.assertEqual(self.paginator.url, next_url)

    def test_normalize_page_data(self):
        """Testing ProxyPaginator.normalize_page_data"""
        proxy = ProxyPaginator(
            self.paginator,
            normalize_page_data_func=lambda data: list(reversed(data)))

        self.assertEqual(proxy.page_data, [3, 2, 1])

    def test_normalize_page_data_on_prev(self):
        """Testing ProxyPaginator.normalize_page_data on prev"""
        proxy = ProxyPaginator(
            self.paginator,
            normalize_page_data_func=lambda data: list(reversed(data)))
        self.paginator.prev_url = 'http://example.com/?page=1'

        data = proxy.prev()

        self.assertEqual(data, [3, 2, 1])

    def test_normalize_page_data_on_next(self):
        """Testing ProxyPaginator.normalize_page_data on next"""
        proxy = ProxyPaginator(
            self.paginator,
            normalize_page_data_func=lambda data: list(reversed(data)))
        self.paginator.next_url = 'http://example.com/?page=3'

        data = proxy.next()

        self.assertEqual(data, [3, 2, 1])



# rbssh plugin for Bazaar.
# Copyright (C) 2017  Beanbag, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""rbssh plugin for Bazaar and Breezy.

This registers :command:`rbssh` as a SSH plugin for Bazaar. It's run entirely
outside of the Review Board process, and is invoked exclusively by
:command:`bzr` or :command:`brz`.
"""

# NOTE: This is a plugin compatible with both Breezy and Bazaar, which is
#       called as an external process. This plugin runs in Breezy/Bazaar's
#       process, which might even be outside of the same Python environment
#       as Review Board. As such, it needs to remain compatible with Python
#       2 and 3, and cannot import any modules from Review Board.
from __future__ import unicode_literals

try:
    from breezy.transport.ssh import (SubprocessVendor,
                                      register_default_ssh_vendor,
                                      register_ssh_vendor)
except ImportError:
    from bzrlib.transport.ssh import (SubprocessVendor,
                                      register_default_ssh_vendor,
                                      register_ssh_vendor)


class RBSSHVendor(SubprocessVendor):
    """SSH vendor class for using rbssh."""

    executable_path = 'rbssh'

    def _get_vendor_specific_argv(self, username, host, port=None,
                                  subsystem=None, command=None):
        """Return arguments to pass to rbssh.

        Args:
            username (unicode):
                The username to connect with.

            host (unicode):
                The hostname to connect to.

            port (int, optional):
                The custom port to connect to.

            subsystem (unicode, optional):
                The SSH subsystem to use.

            command (unicode, optional):
                The command to invoke through the SSH connection.

        Returns:
            list of unicode:
            The list of arguments to pass to :command:`rbssh`.
        """
        args = [self.executable_path]

        if port is not None:
            args.extend(['-p', '%s' % port])

        if username is not None:
            args.extend(['-l', username])

        if subsystem is not None:
            args.extend(['-s', host, subsystem])
        else:
            args.extend([host] + command)

        return args


vendor = RBSSHVendor()
register_ssh_vendor('rbssh', vendor)
register_default_ssh_vendor(vendor)



"""Unit tests for reviewboard.actions.renderers.BaseActionGroupRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.renderers import (BaseActionGroupRenderer,
                                           ButtonActionRenderer,
                                           DefaultActionGroupRenderer,
                                           MenuActionGroupRenderer)
from reviewboard.actions.tests.base import (TestActionsRegistry,
                                            TestGroupAction,
                                            TestGroupActionWithSubgroups,
                                            TestGroupItemAction1,
                                            TestGroupItemAction2,
                                            TestSubgroupAction)
from reviewboard.testing import TestCase


class BaseActionGroupRendererTests(TestCase):
    """Unit tests for BaseActionGroupRenderer.

    Version Added:
        8.0
    """

    def test_get_js_view_data(self) -> None:
        """Testing BaseActionGroupRenderer.get_js_view_data"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        renderer = BaseActionGroupRenderer(action=action,
                                           placement=placement)

        self.assertEqual(renderer.get_js_view_data(context=Context()),
                         {})

    def test_get_extra_context(self) -> None:
        """Testing BaseActionGroupRenderer.get_extra_context"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(item_action_1)
        registry.register(item_action_2)

        renderer = BaseActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        # This should just call out to the action's get_extra_context().
        self.assertEqual(
            renderer.get_extra_context(request=request,
                                       context=context),
            {
                'action': action,
                'action_renderer': renderer,
                'attachment_point_id': 'review-request',
                'children': [
                    item_action_1,
                    item_action_2,
                ],
                'dom_element_id': 'action-review-request-group-action',
                'has_parent': False,
                'id': 'group-action',
                'is_toplevel': True,
                'label': 'Test Group',
                'placement': placement,
                'url': '#',
                'verbose_label': None,
                'visible': True,
            })

    def test_render(self) -> None:
        """Testing BaseActionGroupRenderer.render"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(item_action_1)
        registry.register(item_action_2)

        renderer = BaseActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        # There's no template on the base renderer, so this should be empty.
        self.assertIsInstance(html, SafeString)
        self.assertEqual(html, '')

    def test_render_with_subgroup_renderer(self) -> None:
        """Testing BaseActionGroupRenderer.render with subclass using
        subgroup renderer
        """
        class MyActionGroupRenderer(DefaultActionGroupRenderer):
            default_item_renderer_cls = ButtonActionRenderer
            default_subgroup_renderer_cls = MenuActionGroupRenderer

        action = TestGroupActionWithSubgroups()
        placement = action.get_placement('header')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()
        subgroup = TestSubgroupAction()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(item_action_1)
        registry.register(item_action_2)
        registry.register(subgroup)

        renderer = MyActionGroupRenderer(action=action,
                                         placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="group">
             <button class="ink-c-button"
                     id="action-header-group-item-1-action"
                     aria-label="Group Item 1"
                     type="button">
              <label class="ink-c-button__label">
               Group Item 1
              </label>
             </button>
             <span id="action-header-subgroup-action"
                   role="presentation">
              <a aria-label="Subgroup"
                 href="#"
                 role="presentation">
               Subgroup
               <span class="ink-i-dropdown"></span>
              </a>
              <div hidden
                   style="display: none;">
               <a id="action-header-group-item-2-action"
                  href="#"
                  role="button">
                Group Item 2
               </a>
              </div>
             </span>
            </li>
            """)

    def test_render_with_subgroup_renderer_self(self) -> None:
        """Testing BaseActionGroupRenderer.render with subclass using self
        as subgroup renderer
        """
        class MyActionGroupRenderer(DefaultActionGroupRenderer):
            default_subgroup_renderer_cls = 'self'

        action = TestGroupActionWithSubgroups()
        placement = action.get_placement('header')

        item_action_1 = TestGroupItemAction1()
        item_action_2 = TestGroupItemAction2()
        subgroup = TestSubgroupAction()

        registry = TestActionsRegistry()
        registry.register(action)
        registry.register(item_action_1)
        registry.register(item_action_2)
        registry.register(subgroup)

        renderer = MyActionGroupRenderer(action=action,
                                         placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render(request=request,
                               context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="group">
             <a id="action-header-group-item-1-action"
                role="button"
                href="#">
              Group Item 1
             </a>
             <a id="action-header-group-item-2-action"
                href="#"
                role="button">
              Group Item 2
             </a>
            </li>
            """)

    def test_render_js(self) -> None:
        """Testing BaseActionGroupRenderer.render_js"""
        action = TestGroupAction()
        placement = action.get_placement('review-request')

        registry = TestActionsRegistry()
        registry.register(action)

        renderer = BaseActionGroupRenderer(action=action,
                                           placement=placement)
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        html = renderer.render_js(request=request,
                                  context=context)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            page.addActionView(new RB.Actions.ActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-group-action'),
                model: page.getAction("group-action"),
            }));
            """)



"""Unit tests for reviewboard.reviews.managers.ReviewRequestManager."""

from __future__ import annotations

import kgb
from django_assert_queries.testing import assert_queries
from django.contrib.auth.models import AnonymousUser, User
from djblets.testing.decorators import add_fixtures

from reviewboard.diffviewer.models import DiffSetHistory
from reviewboard.reviews.models import (DefaultReviewer, ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.reviews.signals import review_request_diffset_uploaded
from reviewboard.reviews.testing.queries.review_requests import (
    get_review_requests_accessible_equeries,
    get_review_requests_from_user_equeries,
    get_review_requests_to_group_equeries,
    get_review_requests_to_or_from_user_equeries,
    get_review_requests_to_user_directly_equeries,
    get_review_requests_to_user_equeries,
    get_review_requests_to_user_groups_equeries,
)
from reviewboard.scmtools.errors import ChangeNumberInUseError
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class ReviewRequestManagerTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.managers.ReviewRequestManager."""

    fixtures = ['test_users']

    @add_fixtures(['test_scmtools'])
    def test_create_with_site(self):
        """Testing ReviewRequest.objects.create with LocalSite"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        review_request = ReviewRequest.objects.create(
            user, repository, local_site=local_site)
        self.assertEqual(review_request.repository, repository)
        self.assertEqual(review_request.local_site, local_site)
        self.assertEqual(review_request.local_id, 1)

    @add_fixtures(['test_scmtools'])
    def test_create_with_site_and_commit_id(self):
        """Testing ReviewRequest.objects.create with LocalSite and commit ID"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        review_request = ReviewRequest.objects.create(
            user, repository,
            commit_id='123',
            local_site=local_site)
        self.assertEqual(review_request.repository, repository)
        self.assertEqual(review_request.commit_id, '123')
        self.assertEqual(review_request.local_site, local_site)
        self.assertEqual(review_request.local_id, 1)

    @add_fixtures(['test_scmtools'])
    def test_create_with_site_and_commit_id_conflicts_review_request(self):
        """Testing ReviewRequest.objects.create with LocalSite and commit ID
        that conflicts with a review request
        """
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        # This one should be fine.
        ReviewRequest.objects.create(user, repository, commit_id='123',
                                     local_site=local_site)
        self.assertEqual(local_site.review_requests.count(), 1)

        # This one will yell.
        with self.assertRaises(ChangeNumberInUseError):
            ReviewRequest.objects.create(
                user,
                repository,
                commit_id='123',
                local_site=local_site)

        # Make sure that entry doesn't exist in the database.
        self.assertEqual(local_site.review_requests.count(), 1)

    @add_fixtures(['test_scmtools'])
    def test_create_with_site_and_commit_id_conflicts_draft(self):
        """Testing ReviewRequest.objects.create with LocalSite and
        commit ID that conflicts with a draft
        """
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        # This one should be fine.
        existing_review_request = ReviewRequest.objects.create(
            user, repository, local_site=local_site)
        existing_draft = ReviewRequestDraft.create(existing_review_request)
        existing_draft.commit_id = '123'
        existing_draft.save()

        self.assertEqual(local_site.review_requests.count(), 1)

        # This one will yell.
        with self.assertRaises(ChangeNumberInUseError):
            ReviewRequest.objects.create(
                user,
                repository,
                commit_id='123',
                local_site=local_site)

        # Make sure that entry doesn't exist in the database.
        self.assertEqual(local_site.review_requests.count(), 1)

    @add_fixtures(['test_scmtools'])
    def test_create_with_site_and_commit_id_and_fetch_problem(self):
        """Testing ReviewRequest.objects.create with LocalSite and
        commit ID with problem fetching commit details
        """
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.create(name='test')
        repository = self.create_repository()

        self.assertEqual(local_site.review_requests.count(), 0)
        self.assertEqual(DiffSetHistory.objects.count(), 0)
        self.assertEqual(ReviewRequestDraft.objects.count(), 0)

        with self.assertRaises(NotImplementedError):
            ReviewRequest.objects.create(
                user, repository,
                commit_id='123',
                local_site=local_site,
                create_from_commit_id=True)

        # Make sure that entry and related objects don't exist in the database.
        self.assertEqual(local_site.review_requests.count(), 0)
        self.assertEqual(DiffSetHistory.objects.count(), 0)
        self.assertEqual(ReviewRequestDraft.objects.count(), 0)

    @add_fixtures(['test_scmtools'])
    def test_create_with_create_from_commit_id(self):
        """Testing ReviewRequest.objects.create with commit ID and
        create_from_commit_id
        """
        user = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Test')

        review_request = ReviewRequest.objects.create(
            user,
            repository,
            commit_id='123',
            create_from_commit_id=True)
        self.assertEqual(review_request.repository, repository)
        self.assertEqual(review_request.diffset_history.diffsets.count(), 0)
        self.assertEqual(review_request.commit_id, '123')
        self.assertEqual(review_request.changenum, 123)

        draft = review_request.get_draft()
        self.assertIsNotNone(draft)
        self.assertIsNotNone(draft.diffset)
        self.assertEqual(draft.commit_id, '123')

    @add_fixtures(['test_scmtools'])
    def test_create_with_create_from_commit_id_and_default_reviewers(self):
        """Testing ReviewRequest.objects.create with commit ID,
        create_from_commit_id, and default reviewers
        """
        user = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Test')

        default_reviewer = DefaultReviewer.objects.create(
            name='Default Reviewer',
            file_regex='.')
        default_reviewer.repository.add(repository)
        default_reviewer.people.add(user)
        default_reviewer.groups.add(self.create_review_group())

        review_request = ReviewRequest.objects.create(
            user,
            repository,
            commit_id='123',
            create_from_commit_id=True)
        self.assertEqual(review_request.target_people.count(), 0)
        self.assertEqual(review_request.target_groups.count(), 0)

        draft = review_request.get_draft()
        self.assertIsNotNone(draft)
        self.assertEqual(draft.target_people.count(), 1)
        self.assertEqual(draft.target_groups.count(), 1)

    @add_fixtures(['test_scmtools'])
    def test_create_with_create_from_commit_id_diffset_uploaded(self) -> None:
        """Testing ReviewRequest.objects.create with commit ID from a committed
        change and create_from_commit_id=True emits a
        review_request_diffset_uploaded signal
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        user = User.objects.get(username='doc')
        repository = self.create_repository(tool_name='Test')

        review_request = ReviewRequest.objects.create(
            user,
            repository,
            commit_id='123',
            create_from_commit_id=True)
        self.assertEqual(review_request.repository, repository)
        self.assertEqual(review_request.diffset_history.diffsets.count(), 0)
        self.assertEqual(review_request.commit_id, '123')
        self.assertEqual(review_request.changenum, 123)

        draft = review_request.get_draft()
        diffset = draft.diffset
        self.assertIsNotNone(draft)
        self.assertIsNotNone(diffset)
        self.assertEqual(draft.commit_id, '123')
        self.assertSpyCalledWith(
            on_diffset_uploaded,
            diffset=diffset,
            review_request_draft=draft)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)

    @add_fixtures(['test_scmtools'])
    def test_create_with_create_from_commit_id_no_diffset_uploaded(
        self,
    ) -> None:
        """Testing ReviewRequest.objects.create with commit ID from a pending
        change and create_from_commit_id=True does not emit a
        review_request_diffset_uploaded signal
        """
        def on_diffset_uploaded(diffset, review_request_draft, **kwargs):
            pass

        review_request_diffset_uploaded.connect(on_diffset_uploaded)
        self.spy_on(on_diffset_uploaded)

        user = User.objects.get(username='doc')
        repository = self.create_repository(
            tool_name='TestToolSupportsPendingChangeSets')

        review_request = ReviewRequest.objects.create(
            user,
            repository,
            commit_id=123,
            create_from_commit_id=True)
        self.assertEqual(review_request.repository, repository)
        self.assertEqual(review_request.diffset_history.diffsets.count(), 0)
        self.assertEqual(review_request.commit_id, 123)
        self.assertEqual(review_request.changenum, 123)

        draft = review_request.get_draft()
        diffset = draft.diffset
        self.assertIsNotNone(draft)
        self.assertEqual(draft.commit_id, '123')

        # Since this was created from a pending change, there is no diff to
        # fetch from the repository and no diffset has been created.
        self.assertIsNone(diffset)
        self.assertSpyNotCalled(on_diffset_uploaded)

        review_request_diffset_uploaded.disconnect(on_diffset_uploaded)

    def test_public(self) -> None:
        """Testing ReviewRequest.objects.public"""
        user1 = User.objects.get(username='doc')
        user2 = User.objects.get(username='grumpy')

        review_request_1 = self.create_review_request(
            summary='Test 1',
            publish=True,
            submitter=user1)

        review_request_2 = self.create_review_request(
            summary='Test 2',
            submitter=user2)

        review_request_3 = self.create_review_request(
            summary='Test 3',
            status='S',
            public=True,
            submitter=user1)

        review_request_4 = self.create_review_request(
            summary='Test 4',
            status='S',
            public=True,
            submitter=user2)

        review_request_5 = self.create_review_request(
            summary='Test 5',
            status='D',
            public=True,
            submitter=user1)

        review_request_6 = self.create_review_request(
            summary='Test 6',
            status='D',
            submitter=user2)

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check public() with a viewing user and a default status of 'P'.
        equeries = get_review_requests_accessible_equeries(user=user1)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user1)),
                [
                    review_request_1,
                ])

        # Check public() with anonymous and any status.
        equeries = get_review_requests_accessible_equeries(
            user=AnonymousUser(),
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(status=None)),
                [
                    review_request_5,
                    review_request_4,
                    review_request_3,
                    review_request_1,
                ])

        # Check public() with another viewing user and status=None.
        equeries = get_review_requests_accessible_equeries(
            user=user2,
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user2,
                                                  status=None)),
                [
                    review_request_6,
                    review_request_5,
                    review_request_4,
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

        # Check public() with anonymous and status=None and showing all
        # unpublished changes.
        equeries = get_review_requests_accessible_equeries(
            user=AnonymousUser(),
            show_all_unpublished=True,
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(status=None,
                                                  show_all_unpublished=True)),
                [
                    review_request_6,
                    review_request_5,
                    review_request_4,
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_repository_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public with repository on a
        Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        repository = self.create_repository(
            local_site=local_site)

        review_request = self.create_review_request(
            repository=repository,
            local_site=local_site,
            publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_profile()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True,
            accessible_repository_ids=[repository.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_without_private_repo_access(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repositories
        """
        user = User.objects.get(username='grumpy')

        repository = self.create_repository(
            public=False)

        review_request = self.create_review_request(
            repository=repository,
            publish=True)
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [])

    @add_fixtures(['test_scmtools'])
    def test_public_without_private_repo_access_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repositories on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        repository = self.create_repository(
            public=False,
            local_site=local_site)

        review_request = self.create_review_request(
            repository=repository,
            local_site=local_site,
            publish=True)
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_profile()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_access(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        repositories
        """
        user = User.objects.get(username='grumpy')

        repository = self.create_repository(
            public=False,
            users=[user])

        review_request = self.create_review_request(
            repository=repository,
            publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            accessible_repository_ids=[repository.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_access_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        repositories on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        repository = self.create_repository(
            public=False,
            local_site=local_site,
            users=[user])

        review_request = self.create_review_request(
            repository=repository,
            publish=True,
            local_site=local_site)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_profile()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True,
            accessible_repository_ids=[repository.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_access_through_group(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        repositories
        """
        user = User.objects.get(username='grumpy')

        group = self.create_review_group(
            invite_only=True,
            users=[user])

        repository = self.create_repository(
            public=False,
            review_groups=[group])

        review_request = self.create_review_request(
            repository=repository,
            publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            accessible_repository_ids=[repository.pk],
            accessible_review_group_ids=[group.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_access_through_group_on_local_site(
        self,
    ) -> None:
        """Testing ReviewRequest.objects.public with access to private
        repositories on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        group = self.create_review_group(
            invite_only=True,
            users=[user])

        repository = self.create_repository(
            public=False,
            local_site=local_site,
            review_groups=[group])

        review_request = self.create_review_request(
            repository=repository,
            local_site=local_site,
            publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_profile()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True,
            accessible_repository_ids=[repository.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    def test_public_without_private_group_access(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        group
        """
        user = User.objects.get(username='grumpy')
        group = self.create_review_group(invite_only=True)

        review_request = self.create_review_request(
            publish=True,
            target_groups=[group])
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [])

    def test_public_with_private_group_access(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        group
        """
        user = User.objects.get(username='grumpy')

        group = self.create_review_group(
            invite_only=True,
            users=[user])

        review_request = self.create_review_request(
            publish=True,
            target_groups=[group])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            accessible_review_group_ids=[group.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    def test_public_with_private_group_access_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        group on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        group = self.create_review_group(
            invite_only=True,
            local_site=local_site,
            users=[user])

        review_request = self.create_review_request(
            publish=True,
            local_site=local_site,
            target_groups=[group])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_profile()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True,
            accessible_review_group_ids=[group.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_and_public_group(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repositories and with access to private group
        """
        user = User.objects.get(username='grumpy')
        group = self.create_review_group()

        repository = self.create_repository(public=False)

        review_request = self.create_review_request(
            repository=repository,
            publish=True,
            target_groups=[group])
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            accessible_review_group_ids=[group.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_group_and_public_repo(self) -> None:
        """Testing ReviewRequest.objects.public with access to private
        group and without access to private group
        """
        user = User.objects.get(username='grumpy')
        group = self.create_review_group(invite_only=True)

        repository = self.create_repository(
            public=False,
            users=[user])

        review_request = self.create_review_request(
            repository=repository,
            publish=True,
            target_groups=[group])
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            accessible_repository_ids=[repository.pk])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_and_owner(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repository and as the submitter
        """
        user = User.objects.get(username='grumpy')

        repository = self.create_repository(public=False)
        review_request = self.create_review_request(repository=repository,
                                                    submitter=user,
                                                    publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_and_owner_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repository and as the submitter on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        repository = self.create_repository(
            public=False,
            local_site=local_site)

        review_request = self.create_review_request(
            repository=repository,
            submitter=user,
            local_site=local_site,
            publish=True)
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    def test_public_with_private_group_and_owner(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        group and as the submitter
        """
        user = User.objects.get(username='grumpy')
        group = self.create_review_group(invite_only=True)

        review_request = self.create_review_request(
            submitter=user,
            publish=True,
            target_groups=[group])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    def test_public_with_private_group_and_owner_on_local_site(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        group and as the submitter on a Local Site
        """
        user = User.objects.get(username='grumpy')

        local_site = self.create_local_site(
            name='test',
            users=[user])

        group = self.create_review_group(
            invite_only=True,
            local_site=local_site)

        review_request = self.create_review_request(
            submitter=user,
            local_site=local_site,
            publish=True,
            target_groups=[group])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    @add_fixtures(['test_scmtools'])
    def test_public_with_private_repo_and_target_people(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        repository and user in target_people
        """
        user = User.objects.get(username='grumpy')

        repository = self.create_repository(public=False)

        review_request = self.create_review_request(
            repository=repository,
            publish=True,
            target_people=[user])
        self.assertFalse(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [])

    def test_public_with_private_group_and_target_people(self) -> None:
        """Testing ReviewRequest.objects.public without access to private
        group and user in target_people
        """
        user = User.objects.get(username='grumpy')
        group = self.create_review_group(invite_only=True)

        review_request = self.create_review_request(
            publish=True,
            target_groups=[group],
            target_people=[user])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        equeries = get_review_requests_accessible_equeries(user=user)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user)),
                [
                    review_request,
                ])

    def test_public_with_private_group_and_target_people_on_local_site(
        self,
    ) -> None:
        """Testing ReviewRequest.objects.public without access to private
        group and user in target_people on a Local Site
        """
        local_site = LocalSite.objects.create(name='test')
        user = User.objects.get(username='grumpy')
        local_site.users.add(user)

        group = self.create_review_group(
            invite_only=True,
            local_site=local_site)

        review_request = self.create_review_request(
            publish=True,
            local_site=local_site,
            target_groups=[group],
            target_people=[user])
        self.assertTrue(review_request.is_accessible_by(user))

        # Prime the caches.
        LocalSite.objects.has_local_sites()
        user.get_site_profile(local_site=local_site)

        equeries = get_review_requests_accessible_equeries(
            user=user,
            local_site=local_site,
            has_local_sites_in_db=True)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.public(user=user,
                                                  local_site=local_site)),
                [
                    review_request,
                ])

    def test_to_group(self) -> None:
        """Testing ReviewRequest.objects.to_group"""
        user1 = User.objects.get(username='doc')

        group1 = self.create_review_group(
            name='privgroup',
            users=[user1])

        review_request_1 = self.create_review_request(
            summary='Test 1',
            public=True,
            submitter=user1,
            target_groups=[group1])

        self.create_review_request(
            summary='Test 2',
            public=False,
            submitter=user1,
            target_groups=[group1])

        review_request_3 = self.create_review_request(
            summary='Test 3',
            public=True,
            status='S',
            submitter=user1,
            target_groups=[group1])

        # Check to_group() with a default status of 'P'.
        equeries = get_review_requests_to_group_equeries(
            user=AnonymousUser(),
            to_group_name='privgroup')

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_group('privgroup',
                                                    local_site=None)),
                [
                    review_request_1,
                ])

        # Check to_group() with any status.
        equeries = get_review_requests_to_group_equeries(
            user=AnonymousUser(),
            to_group_name='privgroup',
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_group('privgroup',
                                                    local_site=None,
                                                    status=None)),
                [
                    review_request_3,
                    review_request_1,
                ])

    def test_to_user_group(self) -> None:
        """Testing ReviewRequest.objects.to_user_groups"""
        user1 = User.objects.get(username='doc')
        user2 = User.objects.get(username='grumpy')

        profile1 = user1.get_profile()
        profile2 = user2.get_profile()

        group1 = self.create_review_group(
            name='group1',
            users=[user1])

        group2 = self.create_review_group(
            name='group2',
            users=[user2])

        review_request_1 = self.create_review_request(
            summary='Test 1',
            public=True,
            submitter=user1,
            target_groups=[group1])

        review_request_2 = self.create_review_request(
            summary='Test 2',
            submitter=user2,
            public=True,
            status='S',
            target_groups=[group1])

        review_request_3 = self.create_review_request(
            summary='Test 3',
            public=True,
            submitter=user2,
            target_groups=[group1, group2])

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check to_user_groups() with a default status of 'P'.
        equeries = get_review_requests_to_user_groups_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1,
            target_groups=[group1])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_groups('doc',
                                                          local_site=None)),
                [
                    review_request_3,
                    review_request_1,
                ])

        # Check to_user_groups() with any status.
        equeries = get_review_requests_to_user_groups_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_groups('doc',
                                                          status=None,
                                                          local_site=None)),
                [
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

        # Check to_user_groups() with a viewing user.
        equeries = get_review_requests_to_user_groups_equeries(
            user=user2,
            to_user='grumpy',
            to_user_profile=profile2,
            target_groups=[group2])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_groups('grumpy',
                                                          user=user2,
                                                          local_site=None)),
                [
                    review_request_3,
                ])

    def test_to_or_from_user(self) -> None:
        """Testing ReviewRequest.objects.to_or_from_user"""
        user1 = User.objects.get(username='doc')
        user2 = User.objects.get(username='grumpy')

        profile1 = user1.get_profile()
        profile2 = user2.get_profile()

        group1 = self.create_review_group(
            name='group1',
            users=[user1])

        group2 = self.create_review_group(
            name='group2',
            users=[user2])

        review_request_1 = self.create_review_request(
            summary='Test 1',
            public=True,
            submitter=user1)

        review_request_2 = self.create_review_request(
            summary='Test 2',
            public=False,
            submitter=user1)

        review_request_3 = self.create_review_request(
            summary='Test 3',
            public=True,
            status='S',
            submitter=user1)

        review_request_4 = self.create_review_request(
            summary='Test 4',
            public=True,
            submitter=user1,
            target_groups=[group1],
            target_people=[user2])

        review_request_5 = self.create_review_request(
            summary='Test 5',
            submitter=user2,
            status='S',
            target_groups=[group1],
            target_people=[user2, user1])

        review_request_6 = self.create_review_request(
            summary='Test 6',
            public=True,
            submitter=user2,
            target_groups=[group1, group2],
            target_people=[user1])

        review_request_7 = self.create_review_request(
            summary='Test 7',
            public=True,
            status='S',
            submitter=user2,
            target_people=[user1])

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check to_or_from_user() with a default status of 'P'.
        equeries = get_review_requests_to_or_from_user_equeries(
            user=AnonymousUser(),
            to_or_from_user='doc',
            to_or_from_user_profile=profile1,
            target_groups=[group1])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_or_from_user('doc',
                                                           local_site=None)),
                [
                    review_request_6,
                    review_request_4,
                    review_request_1,
                ])

        # Check to_or_from_user() with a different user.
        equeries = get_review_requests_to_or_from_user_equeries(
            user=AnonymousUser(),
            to_or_from_user='grumpy',
            to_or_from_user_profile=profile2,
            target_groups=[group2])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_or_from_user('grumpy',
                                                           local_site=None)),
                [
                    review_request_6,
                    review_request_4,
                ])

        # Check to_or_from_user() with any status.
        equeries = get_review_requests_to_or_from_user_equeries(
            user=AnonymousUser(),
            to_or_from_user='doc',
            to_or_from_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_or_from_user('doc',
                                                           status=None,
                                                           local_site=None)),
                [
                    review_request_7,
                    review_request_6,
                    review_request_4,
                    review_request_3,
                    review_request_1,
                ])

        # Check to_or_from_user() with a viewing user and any status.
        equeries = get_review_requests_to_or_from_user_equeries(
            user=user2,
            to_or_from_user='doc',
            to_or_from_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_or_from_user('doc',
                                                           user=user2,
                                                           status=None,
                                                           local_site=None)),
                [
                    review_request_7,
                    review_request_6,
                    review_request_5,
                    review_request_4,
                    review_request_3,
                    review_request_1,
                ])

        # Check to_or_from_user() with a viewing user as to/from user and
        # any status.
        equeries = get_review_requests_to_or_from_user_equeries(
            user=user1,
            to_or_from_user='doc',
            to_or_from_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_or_from_user('doc',
                                                           user=user1,
                                                           status=None,
                                                           local_site=None)),
                [
                    review_request_7,
                    review_request_6,
                    review_request_4,
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

    def test_to_user_directly(self) -> None:
        """Testing ReviewRequest.objects.to_user_directly"""
        user1 = User.objects.get(username='doc')
        user2 = User.objects.get(username='grumpy')

        profile1 = user1.get_profile()

        group1 = self.create_review_group(
            name='group1',
            users=[user1])

        group2 = self.create_review_group(
            name='group2',
            users=[user2])

        self.create_review_request(
            summary='Test 1',
            public=True,
            submitter=user1,
            target_groups=[group1],
            target_people=[user2])

        review_request_2 = self.create_review_request(
            summary='Test 2',
            submitter=user2,
            status='S',
            target_groups=[group1],
            target_people=[user2, user1])

        review_request_3 = self.create_review_request(
            summary='Test 3',
            public=True,
            submitter=user2,
            target_groups=[group1, group2],
            target_people=[user1])

        review_request_4 = self.create_review_request(
            summary='Test 4',
            public=True,
            status='S',
            submitter=user2,
            target_people=[user1])

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check to_user_directly() with a default status of 'P'.
        equeries = get_review_requests_to_user_directly_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_directly('doc',
                                                            local_site=None)),
                [
                    review_request_3,
                ])

        # Check to_user_directly() with any status.
        equeries = get_review_requests_to_user_directly_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1,
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_directly('doc',
                                                            status=None)),
                [
                    review_request_4,
                    review_request_3,
                ])

        # Check to_user_directly() with a viewing user and any status.
        equeries = get_review_requests_to_user_directly_equeries(
            user=user2,
            to_user='doc',
            to_user_profile=profile1,
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user_directly('doc',
                                                            user=user2,
                                                            status=None,
                                                            local_site=None)),
                [
                    review_request_4,
                    review_request_3,
                    review_request_2,
                ])

    def test_from_user(self) -> None:
        """Testing ReviewRequest.objects.from_user"""
        user1 = User.objects.get(username='doc')

        review_request_1 = self.create_review_request(
            summary='Test 1',
            public=True,
            submitter=user1)

        review_request_2 = self.create_review_request(
            summary='Test 2',
            public=False,
            submitter=user1)

        review_request_3 = self.create_review_request(
            summary='Test 3',
            public=True,
            status='S',
            submitter=user1)

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check from_user() with a default status of 'P'.
        equeries = get_review_requests_from_user_equeries(
            user=AnonymousUser(),
            from_user='doc')

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.from_user('doc', local_site=None)),
                [
                    review_request_1,
                ])

        # Check from_user() with any status.
        equeries = get_review_requests_from_user_equeries(
            user=AnonymousUser(),
            status=None,
            from_user='doc')

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.from_user('doc',
                                                     status=None,
                                                     local_site=None)),
                [
                    review_request_3,
                    review_request_1,
                ])

        # Check from_user() with a viewing user and any status.
        equeries = get_review_requests_from_user_equeries(
            user=user1,
            status=None,
            from_user='doc')

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.from_user('doc',
                                                     user=user1,
                                                     status=None,
                                                     local_site=None)),
                [
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

    def test_to_user(self) -> None:
        """Testing ReviewRequest.objects.to_user"""
        user1 = User.objects.get(username='doc')
        user2 = User.objects.get(username='grumpy')

        profile1 = user1.get_profile()

        group1 = self.create_review_group(
            name='group1',
            users=[user1])

        group2 = self.create_review_group(
            name='group2',
            users=[user2])

        review_request_1 = self.create_review_request(
            summary='Test 1',
            publish=True,
            submitter=user1,
            target_groups=[group1])

        review_request_2 = self.create_review_request(
            summary='Test 2',
            submitter=user2,
            status='S',
            target_groups=[group1],
            target_people=[user1, user2])

        review_request_3 = self.create_review_request(
            summary='Test 3',
            publish=True,
            submitter=user2,
            target_groups=[group1, group2],
            target_people=[user1])

        review_request_4 = self.create_review_request(
            summary='Test 4',
            publish=True,
            status='S',
            submitter=user2,
            target_groups=[group1, group2],
            target_people=[user1])

        # Prime the caches.
        LocalSite.objects.has_local_sites()

        # Check to_user() with a default status of 'P'.
        equeries = get_review_requests_to_user_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1,
            target_groups=[group1])

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user('doc', local_site=None)),
                [
                    review_request_3,
                    review_request_1,
                ])

        # Check to_user() with any status.
        equeries = get_review_requests_to_user_equeries(
            user=AnonymousUser(),
            to_user='doc',
            to_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user('doc',
                                                   status=None,
                                                   local_site=None)),
                [
                    review_request_4,
                    review_request_3,
                    review_request_1,
                ])

        # Check to_user() with a viewing user and any status.
        equeries = get_review_requests_to_user_equeries(
            user=user2,
            to_user='doc',
            to_user_profile=profile1,
            target_groups=[group1],
            status=None)

        with assert_queries(equeries):
            self.assertEqual(
                list(ReviewRequest.objects.to_user('doc',
                                                   user=user2,
                                                   status=None,
                                                   local_site=None)),
                [
                    review_request_4,
                    review_request_3,
                    review_request_2,
                    review_request_1,
                ])

    def assertValidSummaries(self, review_requests, summaries):
        r_summaries = [r.summary for r in review_requests]

        for summary in r_summaries:
            self.assertIn(summary, summaries,
                          'summary "%s" not found in summary list'
                          % summary)

        for summary in summaries:
            self.assertIn(summary, r_summaries,
                          'summary "%s" not found in review request list'
                          % summary)



"""Base support for certificate storage backends.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Generic, TYPE_CHECKING, TypeVar

from django.utils.text import slugify
from typing_extensions import TypedDict

from reviewboard.certs.cert import CertDataFormat, CertPurpose
from reviewboard.certs.errors import CertificateNotFoundError
from reviewboard.certs.utils import normalize_cert_hostname

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping

    from typelets.django.strings import StrOrPromise

    from reviewboard.certs.cert import (Certificate,
                                        CertificateBundle,
                                        CertificateFingerprints)
    from reviewboard.site.models import LocalSite, AnyOrAllLocalSites


logger = logging.getLogger(__name__)


class BaseStoredData(ABC):
    """Base class for a stored certificate-related data.

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The Local Site owning this stored certificate.
    local_site: LocalSite | None

    #: The storage backend managing this certificate.
    storage: BaseCertificateStorageBackend

    #: A unique ID for this data in the storage backend.
    #:
    #: This should be considered an opaque value outside of the storage
    #: backend.
    storage_id: str

    def __init__(
        self,
        *,
        storage: BaseCertificateStorageBackend,
        storage_id: (str | None) = None,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Initialize the stored data.

        Args:
            storage (BaseCertificateStorageBackend):
                The storage backend managing this stored certificate.

            storage_id (str, optional):
                The opaque ID of the stored data in the backend.

                If not provided, :py:meth:`build_storage_id` will be called
                to generate a new ID.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning this stored certificate.
        """
        self.storage = storage
        self.local_site = local_site

        # Use the storage ID if provided, or build a new one off of computed
        # state if not.
        self.storage_id = storage_id or self.build_storage_id()

    def build_storage_id(self) -> str:
        """Return a new storage ID for the stored data.

        Version Added:
            8.0

        Returns:
            str:
            The new storage ID.
        """
        raise NotImplementedError


class BaseStoredCertificate(BaseStoredData):
    """Base class for a stored certificate.

    These represent a certificate stored in the backend, associating
    certificate data with backend-specific data such as a storage ID.

    Backends must subclass this in order to implement the following:

    * :py:meth:`get_cert_file_path`
    * :py:meth:`get_key_file_path`
    * :py:meth:`load_certificate`

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The hostname associated with the certificate.
    #:
    #: Version Added:
    #:     8.0
    hostname: str

    #: The port associated with the certificate.
    #:
    #: Version Added:
    #:     8.0
    port: int

    #: The purpose set for a certificate.
    #:
    #: This defines whether the certificate is used for trusting a remote
    #: server or authenticating Review Board with a service.
    #:
    #: Version Added:
    #:     8.0
    purpose: CertPurpose

    #: The certificate data being stored.
    _certificate: Certificate | None

    def __init__(
        self,
        *,
        storage: BaseCertificateStorageBackend,
        certificate: (Certificate | None) = None,
        hostname: (str | None) = None,
        port: (int | None) = None,
        purpose: (CertPurpose | None) = None,
        storage_id: (str | None) = None,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Initialize the stored certificate.

        Version Changed:
            8.0:
            Added the ``purpose``, ``hostname``, and ``port`` arguments.

        Args:
            storage (BaseCertificateStorageBackend):
                The storage backend managing this stored certificate.

            certificate (reviewboard.certs.cert.Certificate, optional):
                The certificate data being stored.

            hostname (str, optional):
                The hostname associated with the certificate.

                If ``certificate`` is set, this must either match or not be
                provided. If it's not set, then this must be provided.

                Version Added:
                    8.0

            port (int, optional):
                The port associated with the certificate.

                If ``certificate`` is set, this must either match or not be
                provided. If it's not set, then this must be provided.

                Version Added:
                    8.0

            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificate.

                This sets whether the certificate should be used to trust
                a remote server or authenticate Review Board.

                If ``certificate`` is set, this must either match or not be
                provided. If it's not set, then this must be provided.

                Version Added:
                    8.0

            storage_id (str, optional):
                The opaque ID of the stored data in the backend.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning this stored certificate.

        Raises:
            ValueError:
                One or more arguments were not provided or contained invalid
                values.
        """
        if certificate:
            if not hostname:
                hostname = normalize_cert_hostname(certificate.hostname)
            else:
                hostname = normalize_cert_hostname(hostname)

                if normalize_cert_hostname(certificate.hostname) != hostname:
                    raise ValueError(
                        'The provided hostname= argument must match the '
                        'hostname of the provided certificate=.'
                    )

            if not port:
                port = certificate.port
            elif certificate.port != port:
                raise ValueError(
                    'The provided port= argument must match the port '
                    'of the provided certificate=.'
                )

            if not purpose:
                purpose = certificate.purpose
            elif certificate.purpose != purpose:
                raise ValueError(
                    'The provided purpose= argument must match the purpose '
                    'of the provided certificate=.'
                )
        else:
            if not hostname:
                raise ValueError(
                    'Either certificate= or hostname= must be provided.'
                )

            if not port:
                raise ValueError(
                    'Either certificate= or port= must be provided.'
                )

            if not purpose:
                raise ValueError(
                    'Either certificate= or purpose= must be provided.'
                )

            hostname = normalize_cert_hostname(hostname)

        self.hostname = hostname
        self.port = port
        self.purpose = purpose
        self._certificate = certificate

        super().__init__(storage=storage,
                         storage_id=storage_id,
                         local_site=local_site)

    @property
    def certificate(self) -> Certificate:
        """The Certificate data in storage.

        Type:
            reviewboard.certs.cert.Certificate

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the certificate from storage.

                Details are in the error message.
        """
        if self._certificate is None:
            self._certificate = self.load_certificate()

        return self._certificate

    @abstractmethod
    def get_cert_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str:
        """Return the filesystem path to a certificate.

        This must be implemented by subclasses.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested certificate data file format.

        Returns:
            str:
            The resulting file path.
        """
        raise NotImplementedError

    @abstractmethod
    def get_key_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str | None:
        """Return the filesystem path to a certificate private key.

        Not all certificates will have an associated private key. If one is
        not available, this will return ``None``.

        This must be implemented by subclasses.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested private key data file format.

        Returns:
            str:
            The resulting file path, or ``None`` if a private key is not
            available.
        """
        raise NotImplementedError

    @abstractmethod
    def load_certificate(self) -> Certificate:
        """Load and return a Certificate from storage.

        This must be implemented by subclasses.

        Returns:
            reviewboard.certs.cert.Certificate:
            The resulting loaded certificate.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the certificate from storage.

                Details are in the error message.
        """
        raise NotImplementedError


class BaseStoredCertificateBundle(BaseStoredData):
    """Base class for a stored CA bundle.

    These represent a root CA bundle in the backend, associating root and
    intermediary certificate data with backend-specific data such as a
    storage ID.

    Backends must subclass this in order to implement the following:

    * :py:meth:`load_bundle`

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The associated name of the CA bundle.
    #:
    #: This is in :term:`slug` format.
    #:
    #: Version Added:
    #:     8.0
    name: str

    #: The certificate bundle data being stored.
    _bundle: CertificateBundle | None

    def __init__(
        self,
        *,
        storage: BaseCertificateStorageBackend,
        name: (str | None) = None,
        bundle: (CertificateBundle | None) = None,
        storage_id: (str | None) = None,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Initialize the stored certificate.

        Version Changed:
            8.0:
            * Added the ``name`` argument.

        Args:
            storage (BaseCertificateStorageBackend):
                The storage backend managing this stored certificate.

            name (str, optional):
                The associated name of the CA bundle.

                If ``bundle`` is set, this must either match or not be
                provided. If it's not set, then this must be provided.

                Version Added:
                    8.0

            bundle (reviewboard.certs.cert.CertificateBundle, optional):
                The certificate bundle data being stored.

                Either this or ``name`` must be provided.

            storage_id (str, optional):
                The opaque ID of the stored data in the backend.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning this stored certificate.

        Raises:
            ValueError:
                One or more arguments were invalid or missing.
        """
        if bundle:
            if not name:
                name = bundle.name
            elif bundle.name != name:
                raise ValueError(
                    'The provided name= argument must match the name '
                    'of the provided bundle=.'
                )
        else:
            if not name:
                raise ValueError('Either bundle= or name= must be provided.')

        if name != slugify(name):
            raise ValueError('The name must be in slug format.')

        self.name = name
        self._bundle = bundle

        super().__init__(storage=storage,
                         storage_id=storage_id,
                         local_site=local_site)

    @property
    def bundle(self) -> CertificateBundle:
        """The CA bundle data in storage.

        Type:
            reviewboard.certs.cert.CertificateBundle

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the CA bundle from storage.

                Details are in the error message.
        """
        if self._bundle is None:
            self._bundle = self.load_bundle()

        return self._bundle

    @abstractmethod
    def get_bundle_file_path(
        self,
        *,
        data_format: CertDataFormat = CertDataFormat.PEM,
    ) -> str:
        """Return the filesystem path to a certificate bundle.

        This must be implemented by subclasses.

        Args:
            data_format (reviewboard.certs.cert.CertDataFormat, optional):
                The requested certificate bundle data file format.

        Returns:
            str:
            The resulting file path.
        """
        raise NotImplementedError

    @abstractmethod
    def load_bundle(self) -> CertificateBundle:
        """Load and return the CA bundle data from storage.

        Returns:
            reviewboard.certs.cert.CertificateBundle:
            The resulting loaded CA bundle.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the CA bundle from storage.

                Details are in the error message.
        """
        raise NotImplementedError


class BaseStoredCertificateFingerprints(BaseStoredData):
    """Base class for stored certificate fingerprints.

    These represent fingerprints that can be used to verify a server-issued
    certificate, associating them with backend-specific data such as a
    storage ID.

    Backends must subclass this in order to implement the following:

    * :py:meth:`load_fingerprints`

    Version Added:
        6.0
    """

    ######################
    # Instance variables #
    ######################

    #: The hostname associated with the certificate.
    #:
    #: Version Added:
    #:     8.0
    hostname: str

    #: The port associated with the certificate.
    #:
    #: Version Added:
    #:     8.0
    port: int

    #: The fingerprints data being stored.
    _fingerprints: CertificateFingerprints | None

    def __init__(
        self,
        *,
        storage: BaseCertificateStorageBackend,
        hostname: str,
        port: int,
        fingerprints: (CertificateFingerprints | None) = None,
        storage_id: (str | None) = None,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Initialize the stored certificate.

        Version Changed:
            8.0:
            * Added the required ``hostname`` and ``port`` arguments.

        Args:
            storage (BaseCertificateStorageBackend):
                The storage backend managing this stored certificate.

            hostname (str):
                The hostname associated with the fingerprints.

                Version Added:
                    8.0

            port (int):
                The port associated with the fingerprints.

                Version Added:
                    8.0

            fingerprints (reviewboard.certs.cert.CertificateFingerprints):
                The certificate fingerprints data being stored.

            storage_id (str, optional):
                The opaque ID of the stored data in the backend.

            local_site (reviewboard.site.models.LocalSite, optional):
                The Local Site owning this stored certificate.
        """
        self.hostname = normalize_cert_hostname(hostname)
        self.port = port
        self._fingerprints = fingerprints

        super().__init__(storage=storage,
                         storage_id=storage_id,
                         local_site=local_site)

    @property
    def fingerprints(self) -> CertificateFingerprints:
        """The certificate fingerprint data in storage.

        Type:
            reviewboard.certs.cert.CertificateFingerprints

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the fingerprints from storage.

                Details are in the error message.
        """
        if self._fingerprints is None:
            self._fingerprints = self.load_fingerprints()

        return self._fingerprints

    @abstractmethod
    def load_fingerprints(self) -> CertificateFingerprints:
        """Load and return the certificate fingerprint data from storage.

        Returns:
            reviewboard.certs.cert.CertificateFingerprints:
            The resulting loaded fingerprints.

        Raises:
            reviewboard.certs.errors.CertificateStorageError:
                There was an error loading the fingerprints from storage.

                Details are in the error message.
        """
        raise NotImplementedError


class StorageStats(TypedDict):
    """Statistics about the content in a storage backend.

    This contains counts for stored CA bundles, certificates, and verified
    fingerprints. Along with this, it contains a UUID representing the current
    generation of the stats, which is useful for caching.

    Version Added:
        6.0
    """

    #: The number of stored CA bundles.
    ca_bundle_count: int

    #: The total number of stored certificates.
    #:
    #: This must be a sum of all the values in
    #: :py:attr:`cert_counts_by_purpose`.
    cert_count: int

    #: The number of stored certificates by purpose.
    #:
    #: Version Added:
    #:     8.0
    cert_counts_by_purpose: Mapping[CertPurpose, int]

    #: The number of stored verified fingerprints.
    fingerprint_count: int

    #: A UUID specific to the current state of the content in the backend.
    #:
    #: This can be used to help with caching and invalidation of other state.
    state_uuid: str


_StoredCertT = TypeVar(
    '_StoredCertT',
    bound=BaseStoredCertificate,
)
_StoredCertBundleT = TypeVar(
    '_StoredCertBundleT',
    bound=BaseStoredCertificateBundle,
)
_StoredCertFingerprintsT = TypeVar(
    '_StoredCertFingerprintsT',
    bound=BaseStoredCertificateFingerprints,
)


class BaseCertificateStorageBackend(
    ABC,
    Generic[_StoredCertT, _StoredCertBundleT, _StoredCertFingerprintsT],
):
    """Base class for a certificate storage backend.

    Store backends are responsible for storing certificate and private key
    data, CA bundles, and fingerprint verification information.

    Backends must be able to persist certificates, private keys, and CA bundles
    on the filesystem for the lifetime of the web server process in order to
    allow consumers to access them on demand.

    Subclasses must implement the following attributes:

    * :py:attr:`backend_id`
    * :py:attr:`name`

    And the following methods:

    * :py:meth:`add_ca_bundle`
    * :py:meth:`add_certificate`
    * :py:meth:`add_fingerprints`
    * :py:meth:`delete_ca_bundle`
    * :py:meth:`delete_certificate`
    * :py:meth:`delete_fingerprints`
    * :py:meth:`get_stats`
    * :py:meth:`get_stored_ca_bundle_by_id`
    * :py:meth:`get_stored_ca_bundle`
    * :py:meth:`get_stored_certificate_by_id`
    * :py:meth:`get_stored_certificate`
    * :py:meth:`get_stored_fingerprints_by_id`
    * :py:meth:`get_stored_fingerprints`
    * :py:meth:`iter_ca_bundles`
    * :py:meth:`iter_stored_ca_bundles`
    * :py:meth:`iter_stored_certificates`
    * :py:meth:`iter_stored_fingerprints`

    Version Changed:
        8.0:
        All certificate storage methods have been updated to take a
        ``purpose`` parameter, which is now considered mandatory.

    Version Added:
        6.0
    """

    #: The ID of this storage backend.
    #:
    #: This must be provided by subclasses, and must be unique.
    backend_id: (str | None) = None

    #: The display name of the storage backend.
    #:
    #: This must be provided by subclasses.
    name: (StrOrPromise | None) = ''

    ######################
    # Instance variables #
    ######################

    #: The path to the local filesystem storage usable by this backend.
    #:
    #: This is set when constructing the backend, and should not be changed.
    #: All filesystem operations must be limited to directories under this
    #: path.
    storage_path: str

    def __init__(
        self,
        storage_path: str,
    ) -> None:
        """Initialize the storage backend."""
        assert self.backend_id, 'backend_id must be set.'
        assert self.name, 'name must be set.'

        self.storage_path = storage_path

    @abstractmethod
    def get_stats(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
    ) -> StorageStats:
        """Return statistics on the certificates managed by the backend.

        This will include the total number of stored certificates, CA bundles,
        and verified fingerprints across zero, one, or all Local Sites.

        It also includes state UUIDs that represent the current states of the
        backend.

        Subclasses must ensure that any modifications to storage result in
        updates to this UUID. They may also want to ensure information is
        properly cached for a period of time, and invalidated upon
        modification.

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite bound to the stats.

                If ``None``, the global site will be used.

        Returns:
            StorageStats:
            The computed or cached storage statistics.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error looking up cetificate stats.
        """
        raise NotImplementedError

    @abstractmethod
    def add_ca_bundle(
        self,
        bundle: CertificateBundle,
        *,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertBundleT:
        """Add a root CA bundle to storage.

        The certificate bundle's name along with ``local_site`` are considered
        unique within the storage backend. If there's an existing bundle with
        this information, it will be replaced.

        Args:
            bundle (reviewboard.certs.cert.CertificateBundle):
                The bundle to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificateBundle:
            The resulting stored certificate bundle.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error storing this CA bundle.
        """
        raise NotImplementedError

    def delete_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Delete a root CA bundle from storage.

        Args:
            name (str):
                The unique name of the CA bundle in storage.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The CA bundle was not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this CA bundle.
        """
        stored_bundle = self.get_stored_ca_bundle(
            name=name,
            local_site=local_site)

        if stored_bundle is None:
            raise CertificateNotFoundError()

        assert stored_bundle.storage_id is not None

        self.delete_ca_bundle_by_id(stored_bundle.storage_id)

    @abstractmethod
    def delete_ca_bundle_by_id(
        self,
        storage_id: str,
    ) -> None:
        """Delete a root CA bundle from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the CA bundle to delete.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this CA bundle.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_ca_bundle(
        self,
        *,
        name: str,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertBundleT | None:
        """Return a root CA bundle in storage.

        Args:
            name (str):
                The unique name of the CA bundle in storage.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificateBundle:
            The stored certificate bundle in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this CA bundle.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_ca_bundle_by_id(
        self,
        storage_id: str,
    ) -> _StoredCertBundleT | None:
        """Return a root CA bundle in storage identified by ID.

        Args:
            storage_id (str):
                The ID of the CA bundle in storage.

        Returns:
            BaseStoredCertificateBundle:
            The stored certificate bundle in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this CA bundle.
        """
        raise NotImplementedError

    @abstractmethod
    def iter_stored_ca_bundles(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
    ) -> Iterator[_StoredCertBundleT]:
        """Iterate through all root CA bundles in storage.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

        Yields:
            BaseStoredCertificateBundle:
            Each stored certificate bundle in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through CA bundles.
        """
        raise NotImplementedError

    def get_ca_bundles_dir(
        self,
        *,
        local_site: (LocalSite | None) = None,
    ) -> str:
        """Return a path containing all CA bundle files.

        Args:
            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the bundles.

                If ``None``, the global site will be used.

        Returns:
            str:
            The path to the bundles directory.
        """
        raise NotImplementedError

    @abstractmethod
    def add_certificate(
        self,
        certificate: Certificate,
        *,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertT:
        """Add a certificate to storage.

        The certificate's hostname and port along with ``local_site`` are
        considered unique within the storage backend. If there's an existing
        certificate with this information, it will be replaced.

        Args:
            certificate (reviewboard.certs.cert.Certificate):
                The certificate to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificate:
            The resulting stored certificate.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error storing this certificate.
        """
        raise NotImplementedError

    def delete_certificate(
        self,
        *,
        hostname: str,
        port: int,
        purpose: CertPurpose,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Delete a certificate from storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificate to delete.

                This indicates whether the certificate to match is used
                to trust a remote server or authenticate Review Board.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this certificate.
        """
        stored_certificate = self.get_stored_certificate(
            hostname=hostname,
            port=port,
            local_site=local_site,
            purpose=purpose,
        )

        if stored_certificate is None:
            raise CertificateNotFoundError()

        assert stored_certificate.storage_id

        self.delete_certificate_by_id(stored_certificate.storage_id)

    @abstractmethod
    def delete_certificate_by_id(
        self,
        storage_id: str,
    ) -> None:
        """Delete a certificate from storage.

        Args:
            storage_id (str):
                The ID of the certificate to delete.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting this certificate.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_certificate(
        self,
        *,
        hostname: str,
        port: int,
        purpose: CertPurpose,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertT | None:
        """Return a certificate from storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificate to fetch.

                This indicates whether the certificate is used to trust a
                remote server or authenticate Review Board.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificate:
            The stored certificate in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_certificate_by_id(
        self,
        storage_id: str,
    ) -> _StoredCertT | None:
        """Return a certificate from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate in storage.

        Returns:
            BaseStoredCertificate:
            The stored certificate in storage, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving this certificate.
        """
        raise NotImplementedError

    @abstractmethod
    def iter_stored_certificates(
        self,
        *,
        purpose: CertPurpose,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
    ) -> Iterator[_StoredCertT]:
        """Iterate through all certificates in storage.

        Version Changed:
            8.0:
            Added the ``purpose`` argument.

        Args:
            purpose (reviewboard.certs.cert.CertPurpose):
                The purpose of the certificates to filter for.

                This will limit results to certificates that match the given
                purpose.

                Version Added:
                    8.0

            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite that owns the certificates.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

        Yields:
            BaseStoredCertificate:
            Each stored certificate in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through stored certificates.
        """
        raise NotImplementedError

    @abstractmethod
    def add_fingerprints(
        self,
        fingerprints: CertificateFingerprints,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertFingerprintsT:
        """Add verified certificate fingerprints to storage.

        The ``hostname``, ``port``, and ``local_site`` values taken together
        must be unique within a storage backend.

        Backends may store this together with the certificates or separately.

        Args:
            hostname (str):
                The hostname serving the certificate.

            port (int):
                The port on the host serving the certificate.

            fingerprints (reviewboard.certs.cert.CertificateFingerprints):
                The certificate fingerprints to add.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificateFingerprints:
            The resulting stored certificate fingerprints.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error storing these fingerprints.
        """
        raise NotImplementedError

    def delete_fingerprints(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> None:
        """Delete certificate fingerprints from storage.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting these fingerprints.
        """
        stored_fingerprints = self.get_stored_fingerprints(
            hostname=hostname,
            port=port,
            local_site=local_site)

        if stored_fingerprints is None:
            raise CertificateNotFoundError()

        assert stored_fingerprints.storage_id is not None

        self.delete_fingerprints_by_id(
            stored_fingerprints.storage_id)

    @abstractmethod
    def delete_fingerprints_by_id(
        self,
        storage_id: str,
    ) -> None:
        """Delete certificate fingerprints from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate fingerprints to delete.

        Raises:
            reviewboard.cert.errors.CertificateNotFoundError:
                The fingerprints were not found.

            reviewboard.cert.errors.CertificateStorageError:
                There was an error deleting these fingerprints.
        """
        raise NotImplementedError

    @abstractmethod
    def iter_stored_fingerprints(
        self,
        *,
        local_site: AnyOrAllLocalSites = None,
        start: int = 0,
    ) -> Iterator[_StoredCertFingerprintsT]:
        """Iterate through all certificate fingerprints in storage.

        Args:
            local_site (reviewboard.site.models.LocalSite or
                        reviewboard.site.models.LocalSite.ALL, optional):
                An optional LocalSite that owns the certificate fingerprints.

                If ``None``, the global site will be used.

            start (int, optional):
                The 0-based index within the list of root CA bundles to start
                iterating at.

        Yields:
            BaseStoredCertificateFingerprints:
            Each stored certificate fingerprints in storage.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error iterating through these fingerprints.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_fingerprints(
        self,
        *,
        hostname: str,
        port: int,
        local_site: (LocalSite | None) = None,
    ) -> _StoredCertFingerprintsT | None:
        """Return certificate fingerprints from storage.

        Args:
            hostname (str):
                The hostname of the certificate in storage.

            port (int):
                The port on the host serving the certificate.

            local_site (reviewboard.site.models.LocalSite, optional):
                An optional LocalSite that owns the certificate.

                If ``None``, the global site will be used.

        Returns:
            BaseStoredCertificateFingerprints:
            The stored certificate fingerprints, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving these fingerprints.
        """
        raise NotImplementedError

    @abstractmethod
    def get_stored_fingerprints_by_id(
        self,
        storage_id: str,
    ) -> _StoredCertFingerprintsT | None:
        """Return certificate fingerpritns from storage identified by ID.

        Args:
            storage_id (str):
                The ID of the certificate fingerprints in storage.

        Returns:
            CertificateFingerprints:
            The stored certificate fingerprints, or ``None`` if not found.

        Raises:
            reviewboard.cert.errors.CertificateStorageError:
                There was an error retrieving these fingerprints.
        """
        raise NotImplementedError



import os
import shutil
import tempfile

import pytest
from djblets.cache.serials import generate_media_serial


@pytest.fixture(autouse=True)
def enable_db_access_for_all_tests(django_db_reset_sequences):
    """Enable database access for all unit tests.

    This is applied to all test functions, ensuring database access isn't
    blocked.
    """
    pass


@pytest.fixture(autouse=True, scope='session')
def setup_siteconfig():
    """Set up the siteconfig for tests.

    This is run at the start of the project-wide test session, putting together
    a suitable test environment for Review Board's test suite.
    """
    from django.conf import settings

    # Default to testing in a non-subdir install.
    settings.SITE_ROOT = '/'

    # Set some defaults for cache serials, in case the tests need them.
    settings.AJAX_SERIAL = 123
    settings.TEMPLATE_SERIAL = 123

    # Set a faster password hasher, for performance.
    settings.PASSWORD_HASHERS = (
        'django.contrib.auth.hashers.SHA1PasswordHasher',
    )

    # Make sure we're using standard static files storage, and not
    # something like Pipeline or S3 (since we don't want to trigger any
    # special behavior). Subclasses are free to override this setting.
    settings.STORAGES = {
        'default': {
            'BACKEND': 'django.core.files.storage.FileSystemStorage',
        },
        'staticfiles': {
            'BACKEND':
                'django.contrib.staticfiles.storage.StaticFilesStorage',
        },
    }

    # By default, don't look up DMARC records when generating From
    # addresses for e-mails. Just assume we can, since we're not
    # sending anything out. Some unit tests will override
    # this.
    settings.DJBLETS_EMAIL_FROM_SPOOFING = 'never'

    # Create a temp directory that tests can rely upon.
    tests_tempdir = tempfile.mkdtemp(prefix='rb-tests-')

    # Configure file paths for static media. This will handle the main
    # static and uploaded media directories, along with extension
    # directories (for projects that need to use them).
    settings.STATIC_URL = settings.SITE_ROOT + 'static/'
    settings.MEDIA_URL = settings.SITE_ROOT + 'media/'
    settings.STATIC_ROOT = os.path.join(tests_tempdir, 'static')
    settings.MEDIA_ROOT = os.path.join(tests_tempdir, 'media')

    settings.SITE_DATA_DIR = os.path.join(tests_tempdir, 'data')
    settings.HAYSTACK_CONNECTIONS['default']['PATH'] = \
        os.path.join(settings.SITE_DATA_DIR, 'search-index')

    required_dirs = [
        settings.SITE_DATA_DIR,
        settings.STATIC_ROOT,
        settings.MEDIA_ROOT,
        os.path.join(settings.MEDIA_ROOT, 'uploaded', 'images'),
        os.path.join(settings.MEDIA_ROOT, 'ext'),
        os.path.join(settings.STATIC_ROOT, 'ext'),
    ]

    for dirname in required_dirs:
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    from django.core import management
    management.call_command('collectstatic',
                            verbosity=0,
                            interactive=False)

    generate_media_serial()

    yield

    shutil.rmtree(tests_tempdir, ignore_errors=True)


@pytest.fixture(scope='session')
def django_db_setup(setup_siteconfig, django_db_setup, django_db_blocker):
    """Set up the Django database.

    This is run at the start of the project-wide test session, setting up
    an initial siteconfig in the database.

    This depends on :py:func:`setup_siteconfig` so that the media and static
    paths point at the test directories before the siteconfig is created.
    Creating it first would store the in-tree paths in the database, and
    those would then be applied back onto the Django settings on the first
    request. pytest 9 no longer runs the autouse session fixture first, so
    the dependency has to be explicit.
    """
    with django_db_blocker.unblock():
        from reviewboard.admin.management.sites import init_siteconfig

        siteconfig = init_siteconfig()
        siteconfig.set('mail_from_spoofing', 'never')
        siteconfig.save(update_fields=('settings',))



from djblets.registries.mixins import ExceptionFreeGetterMixin
from djblets.registries.registry import (
    EntryPointRegistry as DjbletsEntryPointRegistry,
    OrderedRegistry as DjbletsOrderedRegistry,
    Registry as DjbletsRegistry,
    RegistryItemType)


class Registry(ExceptionFreeGetterMixin[RegistryItemType],
               DjbletsRegistry[RegistryItemType]):
    """A registry that does not throw exceptions for failed lookups."""


class EntryPointRegistry(ExceptionFreeGetterMixin[RegistryItemType],
                         DjbletsEntryPointRegistry[RegistryItemType]):
    """A registry that auto-populates from an entry-point."""


class OrderedRegistry(ExceptionFreeGetterMixin[RegistryItemType],
                      DjbletsOrderedRegistry[RegistryItemType]):
    """A registry that keeps track of registration order."""



"""Hooks for working with condition choices.

See :ref:`review-request-condition-choices-hook` for instructions.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import Type

from djblets.conditions.choices import BaseConditionChoice
from djblets.extensions.hooks import (BaseRegistryMultiItemHook,
                                      ExtensionHookPoint)

from reviewboard.reviews.conditions import review_request_condition_choices


class ReviewRequestConditionChoicesHook(
    BaseRegistryMultiItemHook[Type[BaseConditionChoice]],
    metaclass=ExtensionHookPoint,
):
    """Hook to add custom condition choices for review requests.

    See :ref:`review-request-condition-choices-hook` for instructions.

    Version Added:
        8.0
    """

    registry = review_request_condition_choices
